﻿using System;
using System.ComponentModel;
using System.Windows;
using ScaraBattle.Model;
using ScaraBattle.Persistence;
using ScaraBattle.View;
using ScaraBattle.ViewModel;
using Microsoft.Win32;
using System.Windows.Input;

namespace ScaraBattle
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App : Application
    {
        #region Fields

        private GameModel _model;
        private GameViewModel _viewModel;
        private MainWindow _view;

        private SizeDialogBox _sizeDialogBox;

        private bool _over = true;

        #endregion

        #region Constructors

        /// <summary>
        /// Instantiate application
        /// </summary>
        public App()
        {
            Startup += new StartupEventHandler(App_Startup);
        }

        #endregion

        #region Application event handlers

        /// <summary>
        /// Upon startup, this method initializes everything that is needed to run the game
        /// </summary>
        private void App_Startup(object sender, StartupEventArgs e)
        {
            _model = new GameModel(new GameFileDataAccess());
            _model.GameOver += new EventHandler<GameEventArgs>(Model_GameOver);

            _viewModel = new GameViewModel(_model);
            _viewModel.NewGame += new EventHandler(ViewModel_NewGame);
            _viewModel.CreateMap += new EventHandler(ViewModel_CreateMap);
            _viewModel.ExitGame += new EventHandler(ViewModel_ExitGame);
            _viewModel.LoadGame += new EventHandler(ViewModel_LoadGame);
            _viewModel.SaveGame += new EventHandler(ViewModel_SaveGame);
            _viewModel.Help += new EventHandler(ViewModel_RequestHelp);
            _viewModel.HelpTower += new EventHandler(ViewModel_RequestHelpTower);
            _viewModel.HelpUnit += new EventHandler(ViewModel_RequestHelpUnit);
            _viewModel.HelpMapeditor += new EventHandler(ViewModel_RequestHelpMapeditor);
            _viewModel.EditMap += new EventHandler(ViewModel_EditMap);

            EventManager.RegisterClassHandler(typeof(Window), Keyboard.KeyUpEvent, new KeyEventHandler(KeyUp), true);

            _view = new MainWindow();
            _view.DataContext = _viewModel;
            _view.Closing += new CancelEventHandler(View_Closing);
            _view.Show();
        }

        #endregion

        #region View event handlers

        /// <summary>
        /// Event handler for closing the game
        /// </summary>
        private void View_Closing(object sender, CancelEventArgs e)
        {
            if (MessageBox.Show("Biztosan ki szeretne lépni?", "ScaraBattle", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.No)
            {
                e.Cancel = true;
            }
        }

        /// <summary>
        /// Event handler for pressing keys
        /// </summary>
        private void KeyUp(object sender, KeyEventArgs e)
        {
            if (!_over)
            {
                switch (e.Key)
                {
                    case Key.A:
                        _viewModel.OnCameraLeft();
                        break;
                    case Key.S:
                        _viewModel.OnCameraDown();
                        break;
                    case Key.D:
                        _viewModel.OnCameraRight();
                        break;
                    case Key.W:
                        _viewModel.OnCameraUp();
                        break;
                }
            }
        }

        #endregion

        #region ViewModel event handlers

        /// <summary>
        /// Event handler for creating a new game
        /// </summary>
        private void ViewModel_NewGame(object sender, EventArgs e)
        {
            _sizeDialogBox = new SizeDialogBox();

            _sizeDialogBox.ShowDialog();

            if (_sizeDialogBox.DialogResult.HasValue && _sizeDialogBox.DialogResult.Value)
            {
                _over = false;

                int width = _sizeDialogBox.MapWidth;
                int height = _sizeDialogBox.MapHeight;
                int barracks = _sizeDialogBox.Barracks;
                int terrain = _sizeDialogBox.Terrain;

                _model.GenerateMap(width, height, barracks, terrain);
            }
        }

        /// <summary>
        /// Event handler for creating a new map
        /// </summary>
        private void ViewModel_CreateMap(object sender, EventArgs e)
        {
            string mapError = _model.CheckMap();

            if (mapError == String.Empty)
            {
                _model.InitializeFromMapEditor();
            }
            else
            {
                MessageBox.Show(mapError, "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event handler for loading new game
        /// </summary>
        private async void ViewModel_LoadGame(object sender, EventArgs e)
        {
            try
            {
                OpenFileDialog openFileDialog = new OpenFileDialog();
                openFileDialog.Title = "ScaraBattle tábla betöltése";
                openFileDialog.Filter = "ScaraBattle File|*.sbf";
                if (openFileDialog.ShowDialog() == true)
                {
                    await _model.LoadGameAsync(openFileDialog.FileName);
                    if (_model.CheckMap() != string.Empty)
                    {
                        throw new GameDataException();
                    }


                    _over = false;
                }
            }
            catch (GameDataException)
            {
                MessageBox.Show("Játék betöltése sikertelen!\nHibás az elérési út, vagy a fájlformátum.", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event handler for saving game
        /// </summary>
        private async void ViewModel_SaveGame(object sender, EventArgs e)
        {
            if (_over)
            {
                MessageBox.Show("Nincs aktív játék, amit el lehetne menteni!", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            } else if (_model.Table.State == GameState.Simulation )
            {
                MessageBox.Show("Szimuláció közben nem lehet játékot menteni!", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            try
            {
                string mapError = _model.CheckMap();

                if (mapError == String.Empty)
                {
                    SaveFileDialog saveFileDialog = new SaveFileDialog();
                    saveFileDialog.Title = "ScaraBattle tábla betöltése";
                    saveFileDialog.Filter = "ScaraBattle File|*.sbf";
                    if (saveFileDialog.ShowDialog() == true)
                    {
                        try
                        {
                            await _model.SaveGameAsync(saveFileDialog.FileName);
                            MessageBox.Show("Sikeres mentés.", "Mentés", MessageBoxButton.OK, MessageBoxImage.Information);
                        }
                        catch (GameDataException)
                        {
                            MessageBox.Show("Játék mentése sikertelen!\nHibás az elérési út, vagy a könyvtár nem írható.", "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                        }
                    }
                }
                else
                {
                    MessageBox.Show(mapError, "Hiba!", MessageBoxButton.OK, MessageBoxImage.Error);
                }



            }
            catch
            {
                MessageBox.Show("A fájl mentése sikertelen!", "ScaraBattle", MessageBoxButton.OK, MessageBoxImage.Error);
            }
        }

        /// <summary>
        /// Event handler for exiting game
        /// </summary>
        private void ViewModel_ExitGame(object sender, EventArgs e)
        {
            _view.Close(); // ablak bezárása
        }

        /// <summary>
        /// Method for telling the player the general rules of the game
        /// </summary>
        private void ViewModel_RequestHelp(object sender, EventArgs e)
        {
            MessageBox.Show("A játék célja, hogy lerombold az ellenfeled kastályát.\n" +
                "A játék körökből áll, a körök pedig építkezés és támadás fázisból.\n\n" +
                "Építkezés:\nKépezz egységeket (amennyiben van elég aranyad), amik a legrövidebb úton megpróbálnak eljutni " +
                "az ellenfeled kastélyához az akadályokat kikerülve, hogy megsebezzék. " +
                "Építs tornyokat, amik elpusztítják az ellenfeled közeledő egységeit. " +
                "A tornyaidat fejlesztheted vagy le is bonthatod.\n\n" +
                "Támadás:\nAz egységek automatikusan mozognak adott számú mezőt, a " +
                "tornyok tüzet nyitnak a hatókörükben lévő valamely ellenséges egységre, amivel elpusztítják azt. " +
                "Amennyiben egy egység rálép az ellenfél kastélyára, megsemmisül és csökkenti a kastély életerejét, " +
                "ami ha nullára csökkent, akkor veszített a tulajdonosa. Ha egy körben mindkét játékos kastélya összeomlik, " +
                "akkor az eredmény döntetlen. Több egység is tartózkodhat egy mezőn (akár ellentétes játékosoké is), " +
                "az egységek nem bocsátkoznak harcba egymással. \n\n" +
                "Pénzszerzés:\n" +
                "Minden kör elején egy adott összeg + korábban kiképzett egységekért + előző körben elpusztított ellenséges egységekért.\n\n" +
                "Megjelenítés:\n" +
                "Bármely cellára kattintva (ha előtte nincs kiválasztva pl. építeni kívánt torony) " +
                "megjelenik alul a cella tartalma, pl több egység esetén, hogy melyik játékosnak hány egysége tartózkodik ott, " +
                "vagy tornyok esetén a részletek (pl hatótávolság).\n" +
                "A \"kamerát\" a WASD billentyűkkel vagy a pálya 4 oldalán található gombokkal mozgathatod és nézheted meg a pálya éppen nem látható részeit."
                , "Általános információk a játékról",
                MessageBoxButton.OK, MessageBoxImage.Information);
        }


        /// <summary>
        /// Method for telling the player the rules of buying and using towers
        /// </summary>
        private void ViewModel_RequestHelpTower(object sender, EventArgs e)
        {
            MessageBox.Show("A saját körödben vehetsz tornyokat, amennyiben van rá elég pénzed. " +
                "\nFejlesztheted a tornyaidat, amivel növelheted a sebzését és hatótávját. " +
                "Később lebonthatod, ha már nincs szükséged rá, ilyenkor a ráköltött pénzed egy részét visszakapod. " +
                "Az ellenfeled tornyait nem tudod fejleszteni vagy lebontani. " +
                "\n\nA tornyaid segítenek megvédeni a kastélyodat úgy, hogy a hozzájuk éppen közel lévő egységeket " +
                "(amik beleesnek a hatótávba, és azok közül pedig azt, amelyik a kastélyodhoz a leghamarabb érnének el) lövik." +
                " \nPénz szempontjából megéri előbb kis tornyot venni és utána fejleszteni. A köröd kezdetén minden előző körben lelőtt ellenséges egységért kapsz plusz pénzt.", 
                "Torony súgó", MessageBoxButton.OK, MessageBoxImage.Information);
        }

        /// <summary>
        /// Method for telling the player the rules of the units
        /// </summary>
        private void ViewModel_RequestHelpUnit(object sender, EventArgs e)
        {
            MessageBox.Show("Mindkétféle egység ugyanolyan gyorsan megy a pályán. " +
                "\nNem bocsátkoznak harcba egymással, bárhány egység lehet ugyanazon a mezőn (egyformák és különbözők is)." +
                "\n\nKis egység:" +
                "\n20 élete van és 1-et sebezi az ellenfél kastélyát, ha eléri." +
                "\n\nNagy egység:" +
                "\n50 élete van és 2-t sebez." +
                "\nA körödben kapsz plusz pénzt minden előző körben kiképzett egységért." +
                " \nPénz szempontjából megéri előbb kis tornyot venni és utána fejleszteni. A köröd kezdetén minden előző körben lelőtt ellenséges egységért kapsz plusz pénzt.",
                "Egység súgó", MessageBoxButton.OK, MessageBoxImage.Information);
        }

        /// <summary>
        /// Method for telling the player the rules of the map editor
        /// </summary>
        private void ViewModel_RequestHelpMapeditor(object sender, EventArgs e)
        {
            MessageBox.Show("Kedved szerint helyezhetsz le objektumokat, amiket jobb oldalon tudsz kiválasztani. " +
                "A bal felső sarokban látod, hogy mi van éppen kivlasztva; ezt fogod lerakni, ha a pálya egyik mezőjére kattintasz. " +
                "Törölni úgy tudsz, ha üres mezőt raksz le. Az üres mezőt úgy tudod kiválasztani, hogy arra az objektumra kattintasz mégegyszer, " +
                "ami éppen ki van jelölve.\n\n'Semleges' objektumok:\n Dűne, Oázis és Rom\n" +
                "Ezek akadályok, egyik játékoshoz sem kötődnek. Az ilyen mezőkre később nem lesz lehetőség tornyokat építeni, " +
                "és az egységek sem mehetnek át rajtuk. Haználható útvonalak kialakítására és lezárására " +
                "(persze csak úgy, hogy ne legyenek elzárva egymástól a barakkok és a kastélyok).\n\n" +
                "'Saját' objektumok: \nKastély, Barakk\nA barakkoktól fognak indulni a kiképzett egységek az llenfél kastélya felé, " +
                "amit megpróbálnak lerombolni. Először mindig kéket raksz le. Ha pirsat szeretnél (vagy a pirosat váltanád vissza kékre), kattints rá mégegyszer " +
                "(nem muszáj egyből, később is lehet, csak legyen ugyanaz kijelölve).\n\nA lerakás feltételei:\nMindkét játékosnak " +
                "pontosan egy kastélya kell, hogy legyen, és minimum 2 barakkja.\nA térképnek átjárhatónak kell lennie, vagyis minden " +
                "barakkból minden kastélynak elérhetőnek kell lennie.\n\nTérkép betöltése:\nHa korábban elmentettél egy térképet, akkor a " +
                "megfelelő fájlt kiválasztva betöltheted, és szerkesztheted még, vagy egyből játszhatsz is vele.\n" +
                "\nTérkép mentése és Játék indítása:\nCsak akkor tudsz menteni és indítani, hogyha a térképed megfelel a feltételeknek. " +
                "Ha nem, akkor részletes listát kapsz arról, hogy miken kell változtatnod, mi hiányzik, vagy mből van túl sok.",
                "Térképszerkesztő súgó", MessageBoxButton.OK, MessageBoxImage.Information);
        }

        /// <summary>
        /// Method opening the map editor
        /// </summary>
        private void ViewModel_EditMap(object sender, EventArgs e)
        {
            _sizeDialogBox = new SizeDialogBox(true);

            _sizeDialogBox.ShowDialog();

            if (_sizeDialogBox.DialogResult.HasValue && _sizeDialogBox.DialogResult.Value)
            {
                _over = false;

                int width = _sizeDialogBox.MapWidth;
                int height = _sizeDialogBox.MapHeight;

                _model.GenerateMap(width, height);
            }
        }

        #endregion

        #region Model event handlers

        /// <summary>
        /// Event handler for game over state
        /// </summary>
        private void Model_GameOver(object sender, GameEventArgs e)
        {
            _viewModel.RefreshTable();

            if (e.WhoWon == 1) // blue
            {
                MessageBox.Show("A kék játékos nyert!");
            }
            else if (e.WhoWon == 2) // red
            {
                MessageBox.Show("A piros játékos nyert!");
            }
            else // draw
            {
                MessageBox.Show("Döntetlen!");
            }

            _over = true;
        }

        #endregion
    }
}
