﻿using System;
using System.Threading.Tasks;
using ScaraBattle.Persistence;
using System.Collections.Generic;
using System.Numerics;
using System.Diagnostics;

namespace ScaraBattle.Model
{
    /// <summary>
    /// The model contains everything that is essential for running the game
    /// </summary>
    public class GameModel
    {
        #region Fields

        private GameTable _table;

        private readonly IGameDataAccess _dataAccess;

        // private variables used for random map generation
        private Random _r;
        private Perlin _p;
        private bool[,] _availablecells;

        // matrices used in the pathfinding algorithm
        private int[,] _distanceBlue; // distance of a given cell from the blue castle
        private (int Y, int X)[,] _stepRed;
        private int[,] _distanceRed;
        private (int Y, int X)[,] _stepBlue;

        private List<Tuple<int, int >> unitsShot;
        private List<Tuple<int , int>> unitsKilled;

        #endregion

        #region Properties

        public GameTable Table { get { return _table; } }

        #endregion

        #region Events

        /// <summary>
        /// Event for game creation
        /// </summary>
        public event EventHandler<GameEventArgs> GameCreated;

        /// <summary>
        /// Event for game advancements
        /// </summary>
        public event EventHandler<GameEventArgs> GameAdvanced;

        /// <summary>
        /// Event for game over
        /// </summary>
        public event EventHandler<GameEventArgs> GameOver;

        #endregion

        #region Constructors

        /// <summary>
        /// Construct a new model of the game
        /// </summary>
        /// <param name="dataAccess">An object that deals with loading and saving the game state</param>
        public GameModel(IGameDataAccess dataAccess)
        {
            _dataAccess = dataAccess;
            _r = new Random();
            _table = new GameTable(0,0,GameState.Menu);
            unitsShot = new List<Tuple<int, int>>();
            unitsKilled = new List<Tuple<int, int>>();
        }

        #endregion

        #region Public game methods

        /// <summary>
        /// Loading a game state
        /// </summary>
        /// <param name="path">Path of the game state</param>
        public async Task LoadGameAsync(String path)
        {
            if (_dataAccess == null)
                throw new InvalidOperationException("No data access is provided.");

            _table = await _dataAccess.LoadAsync(path);

            OnGameCreated();
        }

        /// <summary>
        /// Save current state of the game
        /// </summary>
        /// <param name="path">Path of the game state</param>
        public async Task SaveGameAsync(String path)
        {
            if (_dataAccess == null)
                throw new InvalidOperationException("No data access is provided.");

            if (_table.State == GameState.MapEditor)
            {
                if (CheckMap() != String.Empty)
                {
                    throw new InvalidOperationException("Invalid map!");
                }
                else
                {
                    int barrackCount = 0;
                    for (int i = 0; i < _table.Height; ++i)
                    {
                        for (int j = 0; j < _table.Width; ++j)
                        {
                            if (_table[i, j] is Castle && ((Castle)_table[i, j]).IsBlue)
                            {
                                _table.BlueCastle = (Castle)_table[i, j];
                                _table.BlueCastle.Position = (i, j);
                            }

                            if (_table[i, j] is Castle && !((Castle)_table[i, j]).IsBlue)
                            {
                                _table.RedCastle = (Castle)_table[i, j];
                                _table.RedCastle.Position = (i, j);
                            }

                            if (_table[i, j] is Barrack)
                            {
                                barrackCount++;
                            }
                        }
                    }

                    barrackCount /= 2;

                    _table.BlueBarracks = new (int, int)[barrackCount];
                    _table.RedBarracks = new (int, int)[barrackCount];

                    int blueIndex = 0;
                    int redIndex = 0;
                    for (int i = 0; i < _table.Height; ++i)
                    {
                        for (int j = 0; j < _table.Width; ++j)
                        {
                            if (_table[i, j] is Barrack && ((Barrack)_table[i, j]).IsBlue)
                            {
                                _table.BlueBarracks[blueIndex] = (i, j);
                                blueIndex++;
                            }

                            if (_table[i, j] is Barrack && !((Barrack)_table[i, j]).IsBlue)
                            {
                                _table.RedBarracks[redIndex] = (i, j);
                                redIndex++;
                            }
                        }
                    }

                    _table.BlueTowerPositions = new List<(int, int)>();
                    _table.RedTowerPositions = new List<(int, int)>();

                    _table.BlueUnits = new List<Unit>();
                    _table.RedUnits = new List<Unit>();

                    Table.BlueUnitsCurrentPositions = new List<List<List<Unit>>>();
                    Table.RedUnitsCurrentPositions = new List<List<List<Unit>>>();
                }
            }

            await _dataAccess.SaveAsync(path, _table);
        }

        /// <summary>
        /// Checks the connectivity between two cells on the gametable
        /// </summary>
        /// <param name="x1">X coordinate of first cell</param>
        /// <param name="y1">Y coordinate of first cell</param>
        /// <param name="x2">X coordinate of second cell</param>
        /// <param name="y2">Y coordinate of second cell</param>
        /// <returns>Whether the two points are connected</returns>
        public bool ConnectivityCheck(int x1, int y1, int x2, int y2)
        {
            List <Tuple< int, int>> list = new List<Tuple< int, int>> ();
            int z;
            list.Clear();
            Tuple<int, int> v = new Tuple<int,int>(x1, y1) ;

            list.Add(v);
            z = 0;
            bool connected = false;

            while (z < list.Count && !connected)
            {
                Tuple<int, int> actualcell = list[z];
                if (actualcell.Item1 == x2 && actualcell.Item2 == y2)
                {
                    connected = true;
                }

                if (actualcell.Item1 - 1 > -1)
                {
                    if (_table[(int)actualcell.Item1 - 1, (int)actualcell.Item2] == null || (actualcell.Item1 - 1 == x2 && actualcell.Item2 == y2))
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 - 1, actualcell.Item2);
                        if (!list.Contains(newcell))
                        {
                            list.Add(newcell);

                        }
                    }
                }
                if (actualcell.Item2 + 1 < _table.Width)
                {
                    if (_table[(int)actualcell.Item1, (int)actualcell.Item2 + 1] == null || (actualcell.Item1 == x2 && actualcell.Item2 + 1 == y2))
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 + 1);
                        if (!list.Contains(newcell))
                        {
                            list.Add(newcell);

                        }
                    }
                }
                if (actualcell.Item2 - 1 > -1)
                {
                    if (_table[(int)actualcell.Item1, (int)actualcell.Item2 - 1] == null || (actualcell.Item1 == x2 && actualcell.Item2 - 1 == y2))
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 - 1);
                        if (!list.Contains(newcell))
                        {
                            list.Add(newcell);

                        }
                    }
                }
                if (actualcell.Item1 + 1 < _table.Height)
                {
                    if (_table[(int)actualcell.Item1 + 1, (int)actualcell.Item2] == null || (actualcell.Item1 + 1 == x2 && actualcell.Item2 == y2))
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 + 1, actualcell.Item2);
                        if (!list.Contains(newcell))
                        {
                            list.Add(newcell);

                        }
                    }
                }
                z++;
            }
            return connected;
        }


        /// <summary>
        /// Generates a random map
        /// </summary>
        /// <param name="width">Width of map</param>
        /// <param name="height">Height of map</param>
        /// <param name="barracks">Number of barracks on the map</param>
        /// <param name="terrain">Type of terrain to be generated</param>
        public void GenerateMap(int width, int height, int barracks, int terrain)
        {
            bool done = true;
            do
            {
                _table = new GameTable(height, width, GameState.Blue);

                _table.BlueBarracks = new (int, int)[barracks];
                _table.BlueUnits = new List<Unit>();
                _table.BlueTowerPositions = new List<(int, int)>();

                _table.RedBarracks = new (int, int)[barracks];
                _table.RedUnits = new List<Unit>();
                _table.RedTowerPositions = new List<(int, int)>();

                Table.BlueUnitsCurrentPositions = new List<List<List<Unit>>>();
                Table.RedUnitsCurrentPositions = new List<List<List<Unit>>>();


                _p = new Perlin();
                _p.RandomGrid();



                done = GenerateTable(barracks, terrain);
            } while (!done);

            OnGameCreated();
        }


        /// <summary>
        /// Creates an empty map
        /// </summary>
        /// <param name="width">Width of map</param>
        /// <param name="height">Height of map</param>
        public void GenerateMap(int width, int height)
        {
            _table = new GameTable(height, width, GameState.MapEditor);

            _table.BlueUnits = new List<Unit>();
            _table.BlueTowerPositions = new List<(int, int)>();

            _table.RedUnits = new List<Unit>();
            _table.RedTowerPositions = new List<(int, int)>();

            Table.BlueUnitsCurrentPositions = new List<List<List<Unit>>>();
            Table.RedUnitsCurrentPositions = new List<List<List<Unit>>>();

            _table.BlueBarracks = new (int, int)[width*height];
            
            _table.RedBarracks = new (int, int)[width*height];
           

            OnGameCreated();
        }

        /// <summary>
        /// In the map editor: places or deletes an object, changes the color of castles and barracks.
        /// In the game: tries to buy a tower
        /// </summary>
        /// <param name="y">Y coordinate of cell</param>
        /// <param name="x">X coordinate of cell</param>
        /// <param name="selected">Type of selected cell</param>
        public void PlaceObject(int y, int x, SelectedObject selected)
        {
            if (_table.State == GameState.MapEditor)
            {
                switch (selected)
                {
                    case SelectedObject.Nothing:
                        _table[y, x] = null;
                        break;
                    case SelectedObject.Dune:
                        _table[y, x] = Dune.Instance;
                        break;
                    case SelectedObject.Oasis:
                        _table[y, x] = Oasis.Instance;
                        break;
                    case SelectedObject.Ruins:
                        _table[y, x] = Ruins.Instance;
                        break;
                    case SelectedObject.Castle:
                        if (_table[y, x] is Castle && ((Castle)_table[y, x]).IsBlue)
                        {
                            _table[y, x] = new Castle() { IsBlue = false };
                        }
                        else if (_table[y, x] is Castle && !((Castle)_table[y, x]).IsBlue)
                        {
                            _table[y, x] = new Castle() { IsBlue = true };
                        }
                        else
                        {
                            // by default: blue
                            _table[y, x] = new Castle() { IsBlue = true };
                        }
                        break;
                    case SelectedObject.Barrack:
                        if (_table[y, x] is Barrack && ((Barrack)_table[y, x]).IsBlue)
                        {
                            ((Barrack)_table[y, x]).IsBlue = false;
                        }
                        else if (_table[y, x] is Barrack && !((Barrack)_table[y, x]).IsBlue)
                        {
                            ((Barrack)_table[y, x]).IsBlue = true;
                        }
                        else
                        {
                            // by default: blue
                            _table[y, x] = new Barrack() { IsBlue = true };

                        }
                        break;
                }
            }
            else if (_table.State == GameState.Blue && _table[y, x] == null)
            {
                bool allconnected = true;
                _table[y, x] = Ruins.Instance;
                for (int i = 0; i < _table.BlueBarracks.Length; i++)
                {
                    allconnected = allconnected && ConnectivityCheck(_table.BlueBarracks[i].Item1, _table.BlueBarracks[i].Item2, _table.RedCastle.Position.Y, _table.RedCastle.Position.X);

                }
                for (int i = 0; i < _table.RedBarracks.Length; i++)
                {
                    allconnected = allconnected && ConnectivityCheck(_table.RedBarracks[i].Item1, _table.RedBarracks[i].Item2, _table.BlueCastle.Position.Y, _table.BlueCastle.Position.X);
                }
                _table[y, x] = null;
                switch (selected)
                {
                    case SelectedObject.SmallTower:
                        if (_table.BlueGold >= SmallTower.Price && allconnected)
                        {
                            _table[y, x] = new SmallTower() { IsBlue = true };
                            _table.BlueGold -= SmallTower.Price;
                            _table.BlueTowerPositions.Add((y, x));
                        }
                        else
                        {
                            if (_table.BlueGold < SmallTower.Price && !allconnected)
                            {
                                OnGameAdvanced("Nincs elég pénzed, és el is zárna egy utat a torony, amit le akarsz rakni, ezért nem veheted meg.");
                            }
                            else if (!allconnected)
                            {
                                OnGameAdvanced("Nem rakhatod ide a tornyot, mert elzárna egy utat. Próbálkozz máshol.");
                            }
                            else if (_table.BlueGold < SmallTower.Price)
                            {
                                OnGameAdvanced("Nincs elég pénzed kis tornyot venni.");
                            }
                        }
                        break;
                    case SelectedObject.MediumTower:
                        if (_table.BlueGold >= MediumTower.Price && allconnected)
                        {
                            _table[y, x] = new MediumTower() { IsBlue = true };
                            _table.BlueGold -= MediumTower.Price;
                            _table.BlueTowerPositions.Add((y, x));
                        }
                        else
                        {
                            if (_table.BlueGold < MediumTower.Price && !allconnected)
                            {
                                OnGameAdvanced("Nincs elég pénzed, és el is zárna egy utat a torony, amit le akarsz rakni, ezért nem veheted meg.");
                            }
                            else if (!allconnected)
                            {
                                OnGameAdvanced("Nem rakhatod ide a tornyot, mert elzárna egy utat. Próbálkozz máshol.");
                            }
                            else if (_table.BlueGold < MediumTower.Price)
                            {
                                OnGameAdvanced("Nincs elég pénzed közepes tornyot venni.");
                            }
                        }
                        break;
                    case SelectedObject.LargeTower:
                        if (_table.BlueGold >= LargeTower.Price && allconnected)
                        {
                            _table[y, x] = new LargeTower() { IsBlue = true };
                            _table.BlueGold -= LargeTower.Price;
                            _table.BlueTowerPositions.Add((y, x));
                        }
                        else
                        {
                            if (_table.BlueGold < LargeTower.Price && !allconnected)
                            {
                                OnGameAdvanced("Nincs elég pénzed, és el is zárna egy utat a torony, amit le akarsz rakni, ezért nem veheted meg.");
                            }
                            else if (!allconnected)
                            {
                                OnGameAdvanced("Nem rakhatod ide a tornyot, mert elzárna egy utat. Próbálkozz máshol.");
                            }
                            else if (_table.BlueGold < LargeTower.Price)
                            {
                                OnGameAdvanced("Nincs elég pénzed nagy tornyot venni.");
                            }
                        }
                        break;
                }
            }
            else if (_table.State == GameState.Red && _table[y, x] == null)
            {
                bool allconnected = true;
                _table[y, x] = Ruins.Instance;
                for (int i = 0; i < _table.BlueBarracks.Length; i++)
                {
                    allconnected = allconnected && ConnectivityCheck(_table.BlueBarracks[i].Item1, _table.BlueBarracks[i].Item2, _table.RedCastle.Position.Y, _table.RedCastle.Position.X);

                }
                for (int i = 0; i < _table.RedBarracks.Length; i++)
                {
                    allconnected = allconnected && ConnectivityCheck(_table.RedBarracks[i].Item1, _table.RedBarracks[i].Item2, _table.BlueCastle.Position.Y, _table.BlueCastle.Position.X);
                }
                _table[y, x] = null;
                switch (selected)
                {
                    case SelectedObject.SmallTower:
                        if (_table.RedGold >= SmallTower.Price && allconnected)
                        {
                            _table[y, x] = new SmallTower() { IsBlue = false };
                            _table.RedGold -= SmallTower.Price;
                            _table.RedTowerPositions.Add((y, x));
                        }
                        else
                        {
                            if (_table.BlueGold < SmallTower.Price && !allconnected)
                            {
                                OnGameAdvanced("Nincs elég pénzed, és el is zárna egy utat a torony, amit le akarsz rakni, ezért nem veheted meg.");
                            }
                            else if (!allconnected)
                            {
                                OnGameAdvanced("Nem rakhatod ide a tornyot, mert elzárna egy utat. Próbálkozz máshol.");
                            }
                            else if (_table.BlueGold < SmallTower.Price)
                            {
                                OnGameAdvanced("Nincs elég pénzed kis tornyot venni.");
                            }
                        }
                        break;
                    case SelectedObject.MediumTower:
                        if (_table.RedGold >= MediumTower.Price && allconnected)
                        {
                            _table[y, x] = new MediumTower() { IsBlue = false };
                            _table.RedGold -= MediumTower.Price;
                            _table.RedTowerPositions.Add((y, x));
                        }
                        else
                        {
                            if (_table.BlueGold < MediumTower.Price && !allconnected)
                            {
                                OnGameAdvanced("Nincs elég pénzed, és el is zárna egy utat a torony, amit le akarsz rakni, ezért nem veheted meg.");
                            }
                            else if (!allconnected)
                            {
                                OnGameAdvanced("Nem rakhatod ide a tornyot, mert elzárna egy utat. Próbálkozz máshol.");
                            }
                            else if (_table.BlueGold < MediumTower.Price)
                            {
                                OnGameAdvanced("Nincs elég pénzed közepes tornyot venni.");
                            }
                        }
                        break;
                    case SelectedObject.LargeTower:
                        if (_table.RedGold >= LargeTower.Price && allconnected)
                        {
                            _table[y, x] = new LargeTower() { IsBlue = false };
                            _table.RedGold -= LargeTower.Price;
                            _table.RedTowerPositions.Add((y, x));
                        }
                        else
                        {
                            if (_table.BlueGold < LargeTower.Price && !allconnected)
                            {
                                OnGameAdvanced("Nincs elég pénzed, és el is zárna egy utat a torony, amit le akarsz rakni, ezért nem veheted meg.");
                            }
                            else if (!allconnected)
                            {
                                OnGameAdvanced("Nem rakhatod ide a tornyot, mert elzárna egy utat. Próbálkozz máshol.");
                            }
                            else if (_table.BlueGold < LargeTower.Price)
                            {
                                OnGameAdvanced("Nincs elég pénzed nagy tornyot venni.");
                            }
                        }
                        break;
                }
            } else if (Table.State != GameState.Simulation)
            {
                OnGameAdvanced("A cella foglalt.");
            }
            OnGameAdvanced();
        }


        /// <summary>
        /// Tries to buy the selected type of unit for the current player
        /// </summary>
        /// <param name="unitType">The type of unit we want to buy</param>
        public void BuyUnit(SelectedObject unitType)
        {
            Unit unit = null;

            switch (unitType)
            {
                case SelectedObject.SmallUnit:
                    unit = new SmallUnit() { IsBlue = _table.State == GameState.Blue };
                    break;
                case SelectedObject.LargeUnit:
                    unit = new LargeUnit() { IsBlue = _table.State == GameState.Blue };
                    break;
            }

            if (unit != null)
            {
                if (_table.State == GameState.Blue)
                {
                    if (unitType == SelectedObject.SmallUnit && _table.BlueGold >= SmallUnit.Price)
                    {
                        _table.BlueGold -= SmallUnit.Price;
                        _table.BlueUnits.Add(unit);
                        Table.BlueSmallUnits++;

                        OnGameAdvanced("Sikeres egységvásárlás");
                    }
                    else if (unitType == SelectedObject.LargeUnit && _table.BlueGold >= LargeUnit.Price)
                    {
                        _table.BlueGold -= LargeUnit.Price;
                        _table.BlueUnits.Add(unit);
                        Table.BlueLargeUnits++;

                        OnGameAdvanced("Sikeres egységvásárlás");
                    }
                    else
                    {
                        OnGameAdvanced("Nincs elég pénzed egységeket venni.");
                    }
                }   
                else if (_table.State == GameState.Red)
                {
                    if (unitType == SelectedObject.SmallUnit && _table.RedGold >= SmallUnit.Price)
                    {
                        _table.RedGold -= SmallUnit.Price;
                        _table.RedUnits.Add(unit);
                        Table.RedSmallUnits++;

                        OnGameAdvanced("Sikeres egységvásárlás");
                    }
                    else if (unitType == SelectedObject.LargeUnit && _table.RedGold >= LargeUnit.Price)
                    {
                        _table.RedGold -= LargeUnit.Price;
                        _table.RedUnits.Add(unit);
                        Table.RedLargeUnits++;

                        OnGameAdvanced("Sikeres egységvásárlás");
                    }
                    else
                    {
                        OnGameAdvanced("Nincs elég pénzed egységeket venni.");
                    }
                }
            }
        }

        /// <summary>
        /// Checks whether the current map is valid. Used in map editor, saving and loading.
        /// </summary>
        /// <returns>The error message if something is wrong with the map, otherwise an empty string</returns>
        public string CheckMap()
        {
            string mapError = String.Empty;

            int blueCastleCount = 0;
            int redCastleCount = 0;

            int blueBarrackCount = 0;
            int redBarrackCount = 0;


            (int Y, int X)[] blueBarracks = new (int, int)[_table.Width * _table.Height];
            (int Y, int X)[] redBarracks = new (int, int)[_table.Width * _table.Height];

            // castle coordinates used in connectivity checking
            int castle1x = 0;
            int castle1y = 0;
            int castle2x = 0;
            int castle2y = 0;

            for (int i = 0; i < _table.Height; ++i)
            {
                for (int j = 0; j < _table.Width; ++j)
                {
                    if (_table[i, j] is Castle)
                    {
                        if (((Castle)_table[i, j]).IsBlue)
                        {
                            blueCastleCount++;
                            castle1x = i;
                            castle1y = j;
                        }
                        else
                        {
                            redCastleCount++;
                            castle2x = i;
                            castle2y = j;
                        }
                    }
                    else if (_table[i, j] is Barrack)
                    {
                        if (((Barrack)_table[i, j]).IsBlue)
                        {
                            blueBarracks[blueBarrackCount] = (i, j);
                            blueBarrackCount++;
                        }
                        else
                        {
                            redBarracks[redBarrackCount] = (i, j);
                            redBarrackCount++;
                        }
                    }
                }
            }

            if (blueCastleCount == 1 && redCastleCount == 1)
            {
                bool allconnected = true;
                for (int i = 0; i < blueBarrackCount; i++)
                {
                    allconnected = allconnected && ConnectivityCheck(blueBarracks[i].Y, blueBarracks[i].X, castle2x, castle2y);
                }
                for (int i = 0; i < redBarrackCount; i++)
                {
                    allconnected = allconnected && ConnectivityCheck(redBarracks[i].Y, redBarracks[i].X, castle1x, castle1y);
                }
                if (!allconnected)
                {
                    mapError += "A térkép nem átjárható\n";
                }
            }


            if (blueCastleCount != 1)
            {
                mapError += "A lehelyezett kék kastélyok száma nem egyenlő eggyel\n";
            }

            if (redCastleCount != 1)
            {
                mapError += "A lehelyezett piros kastélyok száma nem egyenlő eggyel\n";
            }

            if (blueBarrackCount != redBarrackCount)
            {
                mapError += "A két játékos barakkjainak száma nem egyezik meg\n";
            }

            if (blueBarrackCount < 2 || redBarrackCount < 2)
            {
                mapError += "Mindkét játékosnak legalább 2 barakkal kell rendelkeznie\n";
            }

            if (mapError != String.Empty)
            {
                mapError = "A térkép hibáinak listája:\n\n" + mapError;
            }

            return mapError;
        }

        /// <summary>
        /// Initializes the game state after starting the game from the map editor. Called after it is confirmed that the map is valid.
        /// </summary>
        public void InitializeFromMapEditor()
        {
            _table.State = GameState.Blue;

            int barrackCount = 0;
            for (int i = 0; i < _table.Height; ++i)
            {
                for (int j = 0; j < _table.Width; ++j)
                {
                    if (_table[i, j] is Barrack)
                    {
                        barrackCount++;
                    }
                    else if (_table[i, j] is Castle)
                    {
                        if (((Castle)_table[i, j]).IsBlue)
                        {
                            _table.BlueCastle.Position = (i, j);
                        }
                        else
                        {
                            _table.RedCastle.Position = (i, j);
                        }
                    }
                }
            }

            _table.BlueBarracks = new (int, int)[barrackCount / 2];
            _table.RedBarracks = new (int, int)[barrackCount / 2];

            int blueIndex = 0;
            int redIndex = 0;

            for (int i = 0; i < _table.Height; ++i)
            {
                for (int j = 0; j < _table.Width; ++j)
                {
                    if (_table[i, j] is Barrack)
                    {
                        if (((Barrack)_table[i, j]).IsBlue)
                        {
                            _table.BlueBarracks[blueIndex] = (i, j);
                            blueIndex++;
                        }
                        else
                        {
                            _table.RedBarracks[redIndex] = (i, j);
                            redIndex++;
                        }
                    }
                }
            }
            OnGameAdvanced();
        }

        /// <summary>
        /// Tries to sell a tower from the selected cell
        /// </summary>
        /// <param name="y">Y coordinate of cell</param>
        /// <param name="x">X coordinate of cell</param>
        public void SellTower(int y, int x)
        {
            if (_table[y, x] != null && _table[y, x] is Tower)
            {
                Tower t = (Tower)_table[y, x];

                if (_table.State == GameState.Blue)
                {
                    if (t.IsBlue)
                    {
                        // blue player sells their tower
                        _table[y, x] = null;
                        _table.BlueTowerPositions.Remove((y, x));
                        _table.BlueGold += t.SellCost;
                        OnGameAdvanced(true);
                    }
                    else
                    {
                        OnGameAdvanced("Ez nem a te tornyod, nem adhatod el.");
                    }
                }
                else if (_table.State == GameState.Red)
                {
                    if (!t.IsBlue)
                    {
                        // red player sells their tower
                        _table[y, x] = null;
                        _table.RedTowerPositions.Remove((y, x));
                        _table.RedGold += t.SellCost;
                        OnGameAdvanced(true);
                    }
                    else
                    {
                        OnGameAdvanced("Ez nem a te tornyod, nem adhatod el.");
                    }
                }
            }
            OnGameAdvanced();
        }

        /// <summary>
        /// Upgrade a tower on the selected cell
        /// </summary>
        /// <param name="y">Y coordinate of cell</param>
        /// <param name="x">X coordinate of cell</param>
        public void UpgradeTower(int y, int x)
        {
            if (_table[y, x] != null && _table[y, x] is Tower)
            {
                Tower t = (Tower)_table[y, x];

                if (_table.State == GameState.Blue)
                {
                    if (t.IsBlue && _table.BlueGold >= t.UpgradeCost)
                    {
                        // blue player upgrades their tower
                        _table[y, x] = NextLevelTower(t);
                        _table.BlueGold -= t.UpgradeCost;

                        OnGameAdvanced(true, y, x);
                    }
                    else if (!t.IsBlue)
                    {
                        OnGameAdvanced("Ez nem a te tornyod, nem fejlesztheted.");
                    }
                    else if (_table.BlueGold < t.UpgradeCost)
                    {
                        
                        OnGameAdvanced("Nincs elég pénzed a tornyod fejlesztésére.");
                    }
                    
                }
                else if (_table.State == GameState.Red)
                {
                    if (!t.IsBlue && _table.RedGold >= t.UpgradeCost)
                    {
                        // red player upgrades their tower
                        _table[y, x] = NextLevelTower(t);
                        _table.RedGold -= t.UpgradeCost;

                        OnGameAdvanced(true, y, x);
                    }
                    else if (t.IsBlue)
                    {
                        OnGameAdvanced("Ez nem a te tornyod, nem fejlesztheted.");
                    }
                    else if (_table.RedGold < t.UpgradeCost)
                    {
                        OnGameAdvanced("Nincs elég pénzed a tornyod fejlesztésére.");
                    }
                }
                OnGameAdvanced();
            }
        }

        /// <summary>
        /// Step into the next game state
        /// </summary>
        public void StepGameState()
        {
            if (_table.State == GameState.Blue)
            {
                _table.State = GameState.Red;
            }
            else if (_table.State == GameState.Red)
            {
                StartSimulation();
                unitsShot = new List<Tuple<int, int>>();
                unitsKilled = new List<Tuple<int, int>>();
                _table.State = GameState.Simulation;
                
            } else if (_table.State == GameState.Simulation)
            {
                _table.State = GameState.Blue;
            }

            OnGameAdvanced(true);
        }

        /// <summary>
        /// Where should a blue unit step to from the given cell to move towards the red castle
        /// </summary>
        /// <param name="i">Y coordinate of current cell</param>
        /// <param name="j">X coordinate of current cell</param>
        /// <returns>The position where the unit has to move</returns>
        public (int Y, int X) GetBlueStep(int i, int j)
        {
            return _stepBlue[i, j];
        }

        /// <summary>
        /// Where should a red unit step to from the given cell to move towards the blue castle
        /// </summary>
        /// <param name="i">Y coordinate of current cell</param>
        /// <param name="j">X coordinate of current cell</param>
        /// <returns>The position where the unit has to move</returns>
        public (int Y, int X) GetRedStep(int i, int j)
        {
            return _stepRed[i, j];
        }

        /// <summary>
        /// Run a full step of the simulation phase
        /// </summary>
        public void Simulate()
        {
            CalculateNextPlace();
            unitsShot = new List<Tuple<int, int>>();
            unitsKilled = new List<Tuple<int, int>>();
            ShootBlueUnits();
            ShootRedUnits();
            CheckSimulationEnd();

            OnGameAdvanced();
        }

        /// <summary>
        /// Checks whether a cell has both blue and red units on it
        /// </summary>
        /// <param name="y">Y coordinate of cell</param>
        /// <param name="x">X coordinate of cell</param>
        /// <returns>Whether the cell has both blue and red units on it</returns>
        public bool IsTwoUnits(int y, int x)
        {
            return Table.BlueUnitsCurrentPositions[y][x].Count != 0 && Table.RedUnitsCurrentPositions[y][x].Count != 0;
        }

        /// <summary>
        /// Returns the units shot in the current step
        /// </summary>
        /// <returns>The units shot in the current step</returns>
        public List<Tuple<int,int>> GetUnitsShot()
        {
            return unitsShot;
        }

        /// <summary>
        /// Returns the units killed in the current step
        /// </summary>
        /// <returns>The units killed in the current step</returns>
        public List<Tuple<int, int>> GetUnitsKilled()
        {
            return unitsKilled;
        }

        public List<Tuple<int, int>> GetTowerRange(int y, int x)
        {
            List<Tuple<int, int>> result = new List<Tuple<int, int>>();

            Tower t = (Tower)_table[y, x];

            int y0 = (int)Math.Round(y - t.Range);
            int x0 = (int)Math.Round(x - t.Range);
            int y1 = (int)Math.Round(y + t.Range);
            int x1 = (int)Math.Round(x + t.Range);

            
            
            for (int i = y0; i <= y1; i++)
            {
                for (int j = x0; j <= x1; j++)
                {
                    if (i>-1 && i < _table.Height && j >-1 && j<_table.Width)
                    {
                        Tuple<int, int> pos = new Tuple<int,int>(i, j);
                           result.Add(pos);
                    }
                }
            }

            return result;
        }

        #endregion

        #region Private game methods

        /// <summary>
        /// Initializes the variables used in the simulation game state
        /// </summary>
        private void InitializeSimulationMatrix()
        {
            Table.BlueUnitsMaxNum += Table.BlueUnits.Count;
            Table.RedUnitsMaxNum += Table.RedUnits.Count;

            for (int i = 0; i < Table.Height; i++)
            {
                Table.BlueUnitsCurrentPositions.Add(new List<List<Unit>>());
                Table.RedUnitsCurrentPositions.Add(new List<List<Unit>>());
                for (int j = 0; j < Table.Width; j++)
                {
                    Table.BlueUnitsCurrentPositions[i].Add(new List<Unit>());
                    Table.RedUnitsCurrentPositions[i].Add(new List<Unit>());
                }
            }
        }

        /// <summary>
        /// Calculate next coordinates for units
        /// </summary>
        private void CalculateNextPlace()
        {
            // Move blue units
            for (int i = 0; i < Table.BlueUnitsCurrentNum; i++)
            {
                // Previous place
                int x = Table.BlueUnits[i].X;
                int y = Table.BlueUnits[i].Y;

                // Calculate new place
                (int Y, int X) delta = GetBlueStep(y, x);
                int newX = x + delta.X;
                int newY = y + delta.Y;
                
                // Remove unit if it reaches the castle in the next step or refresh the place
                if (GetBlueStep(newY, newX) == (0, 0))
                {
                    if (Table.BlueUnits[i] is LargeUnit)
                    {
                        Table.BlueLargeUnits--;
                    } else
                    {
                        Table.BlueSmallUnits--;
                    }

                    Table.BlueUnitsCurrentPositions[y][x].Remove(Table.BlueUnits[i]);
                    Table.BlueUnits.RemoveAt(i);
                    Table.BlueUnitsCurrentNum--;
                    i--;
                    Table.RedCastle.Health--;
                }
                else
                {
                    Table.BlueUnits[i].X = newX;
                    Table.BlueUnits[i].Y = newY;

                    Table.BlueUnitsCurrentPositions[y][x].Remove(Table.BlueUnits[i]);
                    Table.BlueUnitsCurrentPositions[newY][newX].Add(Table.BlueUnits[i]);
                }
            }

            // Move red units
            for (int i = 0; i < Table.RedUnitsCurrentNum; i++)
            {
                // Previous place
                int x = Table.RedUnits[i].X;
                int y = Table.RedUnits[i].Y;

                // Calculate new place
                (int Y, int X) delta = GetRedStep(y, x);
                int newX = x + delta.X;
                int newY = y + delta.Y;

                // Remove unit if it reaches the castle in the next step or refresh the place
                if (GetRedStep(newY, newX) == (0, 0))
                {
                    if (Table.RedUnits[i] is LargeUnit)
                    {
                        Table.RedLargeUnits--;
                    }
                    else
                    {
                        Table.RedSmallUnits--;
                    }

                    Table.RedUnitsCurrentPositions[y][x].Remove(Table.RedUnits[i]);
                    Table.RedUnits.RemoveAt(i);
                    Table.RedUnitsCurrentNum--;
                    i--;
                    Table.BlueCastle.Health--;
                }
                else
                {
                    Table.RedUnits[i].X = newX;
                    Table.RedUnits[i].Y = newY;

                    Table.RedUnitsCurrentPositions[y][x].Remove(Table.RedUnits[i]);
                    Table.RedUnitsCurrentPositions[newY][newX].Add(Table.RedUnits[i]);
                }
            }


            // Put new blue unit on the table if there are more
            if (Table.BlueUnits.Count != Table.BlueUnitsCurrentNum)
            {
                int n = Table.BlueUnitsCurrentNum;
                (int Y, int X) temp = CalcStartPos(true);

                Table.BlueUnits[n].Y = temp.Y;
                Table.BlueUnits[n].X = temp.X;

                Table.BlueUnitsCurrentPositions[temp.Y][temp.X].Add(Table.BlueUnits[n]);
                Table.BlueUnitsCurrentNum++;
            }

            // Put new red unit on the table if there are more
            if (Table.RedUnits.Count != Table.RedUnitsCurrentNum)
            {
                int n = Table.RedUnitsCurrentNum;
                (int Y, int X) temp = CalcStartPos(false);

                Table.RedUnits[n].Y = temp.Y;
                Table.RedUnits[n].X = temp.X;

                Table.RedUnitsCurrentPositions[temp.Y][temp.X].Add(Table.RedUnits[n]);
                Table.RedUnitsCurrentNum++;
            }
        }

        /// <summary>
        /// Clear the variables used in the simulation phase after the simulation ends
        /// </summary>
        private void ClearLists()
        {
            for (int i = 0; i < Table.Height; i++)
            {
                for (int j = 0; j < Table.Width; j++)
                {
                    Table.BlueUnitsCurrentPositions = new List<List<List<Unit>>>();
                    Table.RedUnitsCurrentPositions = new List<List<List<Unit>>>();
                }
            }

            Table.BlueUnits = new List<Unit>();
            Table.RedUnits = new List<Unit>();
            Table.BlueUnitsCurrentNum = 0;
            Table.RedUnitsCurrentNum = 0;
            Table.BlueDeathBonus = 0;
            Table.RedDeathBonus = 0;
            Table.RedLargeUnits = 0;
            Table.BlueLargeUnits = 0;
            Table.RedSmallUnits = 0;
            Table.BlueSmallUnits = 0;
        }

        /// <summary>
        /// Recognizes whether the simulation ended. If so, we step the game state and clear temporary variables
        /// </summary>
        private void CheckSimulationEnd()
        {
            if (_table.RedUnits.Count == 0 && _table.BlueUnits.Count == 0)
            {
                CheckGameEnd();
                StepGameState();
                CalculateGold();
                ClearLists();
            }
        }

        /// <summary>
        /// Increases the gold of each player based on their performance
        /// </summary>
        private void CalculateGold()
        {
            Table.BlueGold += Table.BlueUnitsMaxNum * 10 + Table.BlueDeathBonus;
            Table.RedGold += Table.RedUnitsMaxNum * 10 + Table.RedDeathBonus;
        }

        /// <summary>
        /// Shoots blue units with red towers
        /// </summary>
        private void ShootBlueUnits()
        {
            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++)
                {
                    if (_table[i, j] is Tower)
                    {
                        Tower t = (Tower)_table[i, j];
                        if (t.IsBlue == false)
                        {
                            Tuple<int, int> unitpos = ShootUnit(i, j, t.Range, true);
                            if (unitpos.Item1 != -1)
                            {
                                int k = 0;
                                while (k < _table.BlueUnits.Count && !(_table.BlueUnits[k].Y == unitpos.Item1 && _table.BlueUnits[k].X == unitpos.Item2))
                                {
                                    k++;

                                }

                                if (k < _table.BlueUnits.Count)
                                {
                                    _table.BlueUnits[k].Health -= t.Damage;
                                    if (_table.BlueUnits[k].Health <= 0)
                                    {
                                        if (Table.BlueUnits[k] is LargeUnit)
                                        {
                                            Table.RedDeathBonus += LargeUnit.Price / 5;
                                            Table.BlueLargeUnits--;
                                        }
                                        else
                                        {
                                            Table.RedDeathBonus += SmallUnit.Price / 5;
                                            Table.BlueSmallUnits--;
                                        }
                                        _table.BlueUnits.RemoveAt(k);
                                        _table.BlueUnitsCurrentNum--;
                                        _table.BlueUnitsCurrentPositions[unitpos.Item1][unitpos.Item2].RemoveAt(0);
                                        unitsKilled.Add(unitpos);
                                    }
                                    else
                                    {
                                        unitsShot.Add(unitpos);
                                    }
                                }
                            }
                        }

                    }
                }
            }
        }

        /// <summary>
        /// Shoots red units with blue towers
        /// </summary>
        private void ShootRedUnits()
        {

            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++)
                {
                    if (_table[i, j] is Tower)
                    {
                        Tower t = (Tower)_table[i, j];
                        if (t.IsBlue == true)
                        {
                            Tuple<int, int> unitpos = ShootUnit(i, j, t.Range, false);
                            if (unitpos.Item1 != -1)
                            {
                                int k = 0;
                                while (k < _table.RedUnits.Count && !(_table.RedUnits[k].Y == unitpos.Item1 && _table.RedUnits[k].X == unitpos.Item2))
                                {
                                    k++;

                                }

                                if (k < _table.RedUnits.Count)
                                {
                                    _table.RedUnits[k].Health -= t.Damage;
                                    if (_table.RedUnits[k].Health <= 0)
                                    {
                                        if (Table.RedUnits[k] is LargeUnit)
                                        {
                                            Table.BlueDeathBonus += LargeUnit.Price / 5;
                                            Table.RedLargeUnits--;
                                        } else
                                        {
                                            Table.BlueDeathBonus += SmallUnit.Price / 5;
                                            Table.RedSmallUnits--;
                                        }
                                        
                                        _table.RedUnits.RemoveAt(k);
                                        _table.RedUnitsCurrentNum--;
                                        _table.RedUnitsCurrentPositions[unitpos.Item1][unitpos.Item2].RemoveAt(0);
                                        unitsKilled.Add(unitpos);
                                    }
                                    else
                                    {
                                        unitsShot.Add(unitpos);
                                    }
                                }


                            }
                        }

                    }
                }
            }
        }

        /// <summary>
        /// Calculates a random starting position near one of the barracks for the placement of a new unit
        /// </summary>
        /// <param name="blue">Whether the unit is blue</param>
        /// <returns>A valid starting position for the unit</returns>
        private (int Y, int X) CalcStartPos(bool blue)
        {
            List<(int Y, int X)> startPost = new List<(int Y, int X)>();
            if (blue)
            {
                foreach ((int Y, int X) b in Table.BlueBarracks)
                {
                    startPost.Add(b);
                }
            }
            else
            {
                foreach ((int Y, int X) b in Table.RedBarracks)
                {
                    startPost.Add(b);
                }
            }
            Random rand = new Random();
            int number = rand.Next(0, startPost.Count);
            int deltaX = rand.Next(-1, 2);

            int deltaY = rand.Next(-1, 2);

            (int Y, int X) newPos = startPost[number];

            while (!IsValidStart((newPos.Y + deltaY, newPos.X + deltaX), blue))
            {
                deltaX = rand.Next(-1, 2);
                deltaY = rand.Next(-1, 2);
            }
            newPos.Y += deltaY;
            newPos.X += deltaX;

            return newPos;
        }

        /// <summary>
        /// Checks whether a starting position is valid for a unit of a given color
        /// </summary>
        /// <param name="newPos">The starting position of the unit</param>
        /// <param name="blue">The color of the unit</param>
        /// <returns>Whether the starting position is valid for the unit</returns>
        private bool IsValidStart((int Y, int X) newPos, bool blue)
        {
            if (blue)
            {
                return newPos.X >= 0 && newPos.X < Table.Width && newPos.Y >= 0 &&
                       newPos.Y < Table.Height && GetBlueStep(newPos.Y, newPos.X) != (0, 0);
            }
            else
            {
                return newPos.X >= 0 && newPos.X < Table.Width && newPos.Y >= 0 && 
                       newPos.Y < Table.Height && GetRedStep(newPos.Y, newPos.X) != (0, 0);
            }
        }

        /// <summary>
        /// Calculates the tower one level above the current one
        /// </summary>
        /// <param name="t">The current tower</param>
        /// <returns>The tower one level above the current one</returns>
        private Tower NextLevelTower(Tower t)
        {
            if (t.Level != 3 && t.Level != 6)
            {
                t.Level++;
                t.Range += 0.2f;
                t.Damage++;
            }
            else if (t.Level == 3)
            {
                t = new MediumTower() { IsBlue = _table.State == GameState.Blue };
            }
            else if (t.Level == 6)
            {
                t = new LargeTower() { IsBlue = _table.State == GameState.Blue };
            }

            return t;
        }

        /// <summary>
        /// Generates a random table
        /// </summary>
        private bool GenerateTable(int barracks, int terrain)
        {
            double min = 1;
            double max = -1;
            int count = _table.Width * _table.Height;

            double dune = 0;
            double oasis = 0;
            double ruin = 0;

            switch (terrain)
            {
                case 0:
                    dune = 0.9;
                    oasis = 0.1;
                    ruin = 0.2;
                    break;
                case 1:
                    dune = 0.8;
                    oasis = 0.2;
                    ruin = 0.3;
                    break;
                case 2:
                    dune = 0.75;
                    oasis = 0.25;
                    ruin = 0.4;
                    break;
            }


            // scale perlin noise into the [0, 1] interval
            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++)
                {
                    double perlin = _p.perlinnoise((double)i / _table.Height * 5, (double)j / _table.Width * 5);
                    if (perlin < min)
                    {
                        min = perlin;
                    }
                    if (perlin > max)
                    {
                        max = perlin;
                    }
                    _table[i, j] = null;
                }
            }

            _availablecells = new bool[_table.Height, _table.Width];
            // Calculate the cells where an object needs to be placed
            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++)
                {
                    double perlin = _p.perlinnoise((double)i / _table.Height * 5, (double)j / _table.Width * 5);
                    perlin = perlin / (max - min) - min / (max - min);
                    if (perlin > dune)
                    {
                        _table[i, j] = Dune.Instance;
                        count--;
                    }
                    if (perlin < oasis)
                    {
                        _table[i, j] = Oasis.Instance;
                        count--;
                    }
                    if (perlin > oasis && perlin < ruin)
                    {
                        Random r = new Random();
                        if (r.Next(2) <= 0.5)
                        {
                            _table[i, j] = Ruins.Instance;
                            count--;
                        }
                    }
                    _availablecells[i, j] = false;
                }
            }

            // largest connected part of the map
            // used for the placement of castles and barracks in order to avoid placement to isolated cells
            int maxavailable = 0;
            List< Tuple<int, int> > list = new List<Tuple<int, int>>();
            List<Tuple<int, int>> maxlist = new List<Tuple<int, int>>();

            int z = 0;
            int zi = 0;
            int zj = 0;


            while (count > 0)
            {
                while (zi < _table.Height && (_table[zi, zj] != null || _availablecells[zi, zj] == true))
                {
                    zj++;
                    if (zj >= _table.Width)
                    {
                        zi++;
                        zj = 0;
                    }
                }

                list.Clear();
                Tuple<int, int> v = new Tuple<int, int>(zi, zj);
                count--;

                list.Add(v);
                z = 0;

                while (z < list.Count)
                {
                    Tuple<int, int> actualcell = list[z];
                    _availablecells[(int)actualcell.Item1, (int)actualcell.Item2] = true;

                    if (actualcell.Item1 - 1 > -1)
                    {
                        if (_table[(int)actualcell.Item1 - 1, (int)actualcell.Item2] == null)
                        {
                            Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 - 1, actualcell.Item2);
                            if (!list.Contains(newcell))
                            {
                                list.Add(newcell);
                                count--;
                            }
                        }
                    }
                    if (actualcell.Item2 + 1 < _table.Width)
                    {
                        if (_table[(int)actualcell.Item1, (int)actualcell.Item2 + 1] == null)
                        {
                            Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 + 1);
                            if (!list.Contains(newcell))
                            {
                                list.Add(newcell);
                                count--;
                            }
                        }
                    }
                    if (actualcell.Item2 - 1 > -1)
                    {
                        if (_table[(int)actualcell.Item1, (int)actualcell.Item2 - 1] == null)
                        {
                            Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 - 1);
                            if (!list.Contains(newcell))
                            {
                                list.Add(newcell);
                                count--;
                            }
                        }
                    }
                    if (actualcell.Item1 + 1 < _table.Height)
                    {
                        if (_table[(int)actualcell.Item1 + 1, (int)actualcell.Item2] == null)
                        {
                            Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 + 1, actualcell.Item2);
                            if (!list.Contains(newcell))
                            {
                                list.Add(newcell);
                                count--;
                            }
                        }
                    }
                    z++;
                }

                if (z - 1 > maxavailable)
                {
                    maxavailable = z - 1;
                    maxlist.Clear();
                    maxlist.AddRange(list);
                }

            }


            // the available cells inside the largest connected section
            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++)
                {
                    _availablecells[i, j] = false;
                }
            }

            count = 0;

            for (int i = 0; i < maxlist.Count; i++)
            {
                _availablecells[(int)maxlist[i].Item1, (int)maxlist[i].Item2] = true;
                count++;
            }

            if (count < _table.Width * _table.Height * 0.5)
            {
                return false;
            }

            // placement of castles inside the largest connected section
            int castle1x = 0;
            int castle1y = 0;
            int castle2x = 0;
            int castle2y = 0;

            _table.BlueCastle = new Castle();

            GetRandomFreeCellLeft(ref castle1x, ref castle1y, 3);
            _table[castle1x, castle1y] = _table.BlueCastle;
            _table.BlueCastle.Position = (castle1x, castle1y);
            _table.BlueCastle.IsBlue = true;
            _availablecells[castle1x, castle1y] = false;

            _table.RedCastle = new Castle();
            GetRandomFreeCellRight(ref castle2x, ref castle2y, 3);
            _table[castle2x, castle2y] = _table.RedCastle;
            _table.RedCastle.Position = (castle2x, castle2y);
            _availablecells[castle2x, castle2y] = false;




            // placement of barracks
            bool done = true;
            int iter = 0;
            do
            {

                done = GetBarracks(castle1x, castle1y, castle2x, castle2y, barracks);
                iter++;

            } while (!done && iter < 50);

            if (!done)
            {
                return false;
            }
            return true;
        }

        /// <summary>
        /// Returns an empty cell randomly left from the given position
        /// </summary>
        /// <param name="x">X coordinate of cell</param>
        /// <param name="y">Y coordinate of cell</param>
        /// <param name="s">Normalization variable</param>
        private void GetRandomFreeCellLeft(ref int x, ref int y, int s)
        {
            do
            {
                x = _r.Next(_table.Height);
                y = _r.Next(_table.Width / s);
            }
            while (_table[x, y] != null || _availablecells[x, y] == false);
        }

        /// <summary>
        /// Returns an empty cell randomly right from the given position
        /// </summary>
        /// <param name="x">X coordinate of cell</param>
        /// <param name="y">Y coordinate of cell</param>
        /// <param name="s">Normalization variable</param>
        private void GetRandomFreeCellRight(ref int x, ref int y, int s)
        {
            do
            {
                x = _r.Next(_table.Height);
                y = _r.Next(_table.Width / s) + _table.Width / s * (s - 1);
            }
            while (_table[x, y] != null || _availablecells[x, y] == false);
        }

        /// <summary>
        /// Initializes the barracks randomly on the map
        /// </summary>
        /// <param name="castle1x">X coordinate of blue castle</param>
        /// <param name="castle1y">Y coordinate of blue castle</param>
        /// <param name="castle2x">X coordinate of red castle</param>
        /// <param name="castle2y">Y coordinate of red castle</param>
        /// <param name="barracks">Number of barracks to be placed</param>
        /// <returns>Whether the placement was successful</returns>
        private bool GetBarracks(int castle1x, int castle1y, int castle2x, int castle2y, int barracks)
        {
            for (int i = 0; i < barracks; i++)
            {
                int x = 0;
                int y = 0;
                int iter = 0;
                bool allconnected;
                do
                {
                    GetRandomFreeCellLeft(ref x, ref y, 2);
                    _table[x, y] = new Barrack() { IsBlue = true };
                    allconnected = true;
                    for (int j = 0; j < i; j++)
                    {
                        allconnected = allconnected && ConnectivityCheck(_table.BlueBarracks[j].Item1, _table.BlueBarracks[j].Item2, castle2x, castle2y);
                        allconnected = allconnected && ConnectivityCheck(_table.RedBarracks[j].Item1, _table.RedBarracks[j].Item2, castle1x, castle1y);

                    }
                    _table[x, y] = null;
                    allconnected = allconnected && ConnectivityCheck(x, y, castle2x, castle2y);
                    iter++;
                } while (!allconnected && iter < 50);

                if (!allconnected)
                {
                    for (int j = 0; j < i; j++)
                    {
                        _table[_table.RedBarracks[j].Item1, _table.RedBarracks[j].Item2] = null;
                        _availablecells[_table.RedBarracks[j].Item1, _table.RedBarracks[j].Item2] = true;
                        _table[_table.BlueBarracks[j].Item1, _table.BlueBarracks[j].Item2] = null;
                        _availablecells[_table.BlueBarracks[j].Item1, _table.BlueBarracks[j].Item2] = true;
                    }
                    return false;

                }
                _table.BlueBarracks[i] = (x, y);
                _table[x, y] = new Barrack() { IsBlue = true };
                _availablecells[x, y] = false;

                iter = 0;
                do
                {
                    GetRandomFreeCellRight(ref x, ref y, 2);
                    _table[x, y] = new Barrack() { IsBlue = false };
                    allconnected = true;
                    for (int j = 0; j <= i; j++)
                    {
                        allconnected = allconnected && ConnectivityCheck(_table.BlueBarracks[j].Item1, _table.BlueBarracks[j].Item2, castle2x, castle2y);

                    }
                    for (int j = 0; j < i; j++)
                    {

                        allconnected = allconnected && ConnectivityCheck(_table.RedBarracks[j].Item1, _table.RedBarracks[j].Item2, castle1x, castle1y);
                    }
                    iter++;
                    _table[x, y] = null;
                    allconnected = allconnected && ConnectivityCheck(x, y, castle1x, castle1y);
                } while (!allconnected && iter < 50);
                if (!allconnected)
                {
                    for (int j = 0; j <= i; j++)
                    {
                        _table[_table.BlueBarracks[j].Item1, _table.BlueBarracks[j].Item2] = null;
                        _availablecells[_table.BlueBarracks[j].Item1, _table.BlueBarracks[j].Item2] = true;
                    }
                    for (int j = 0; j < i; j++)
                    {
                        _table[_table.RedBarracks[j].Item1, _table.RedBarracks[j].Item2] = null;
                        _availablecells[_table.RedBarracks[j].Item1, _table.RedBarracks[j].Item2] = true;
                    }
                    return false;

                }
                _table.RedBarracks[i] = (x, y);
                _table[x, y] = new Barrack() { IsBlue = false };
                _availablecells[x, y] = false;

            }

            return true;
        }

        /// <summary>
        /// Calculate the path blue units have to travel
        /// </summary>
        private void GetPathBlue()
        {

            _distanceBlue = new int[_table.Height, _table.Width];
            _stepRed = new (int, int)[_table.Height, _table.Width];

            List<Tuple<int, int>> list = new List<Tuple<int, int>>();
            int z = 0;
            list.Clear();
            Tuple<int, int> v = new Tuple<int, int>(_table.BlueCastle.Position.Y, _table.BlueCastle.Position.X);

            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++) {
                    _distanceBlue[i, j] = (int)(_table.Width * _table.Height + 1);
                    _stepRed[i, j] = (0, 0);
                }
            }

            list.Add(v);
            _distanceBlue[(int)v.Item1, (int)v.Item2] = 0;

            while (z < list.Count)
            {
                Tuple<int, int> actualcell = list[z];

                if (actualcell.Item1 - 1 > -1)
                {
                    if (_table[(int)actualcell.Item1 - 1, (int)actualcell.Item2] == null && _distanceBlue[(int)actualcell.Item1 - 1, (int)actualcell.Item2] > _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 - 1, actualcell.Item2);
                        _distanceBlue[(int)actualcell.Item1 - 1, (int)actualcell.Item2] = _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepRed[(int)actualcell.Item1 - 1, (int)actualcell.Item2] = (1, 0);
                        list.Add(newcell);

                    }
                }
                if (actualcell.Item2 + 1 < _table.Width)
                {
                    if (_table[(int)actualcell.Item1, (int)actualcell.Item2 + 1] == null && _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2 + 1] > _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 + 1);
                        _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2 + 1] = _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepRed[(int)actualcell.Item1, (int)actualcell.Item2 + 1] = (0, -1);
                        list.Add(newcell);

                    }
                }
                if (actualcell.Item2 - 1 > -1)
                {
                    if (_table[(int)actualcell.Item1, (int)actualcell.Item2 - 1] == null && _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2 - 1] > _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 - 1);
                        _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2 - 1] = _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepRed[(int)actualcell.Item1, (int)actualcell.Item2 - 1] = (0, 1);
                        list.Add(newcell);

                    }
                }
                if (actualcell.Item1 + 1 < _table.Height)
                {
                    if (_table[(int)actualcell.Item1 + 1, (int)actualcell.Item2] == null && _distanceBlue[(int)actualcell.Item1 + 1, (int)actualcell.Item2] > _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 + 1, actualcell.Item2);
                        _distanceBlue[(int)actualcell.Item1 + 1, (int)actualcell.Item2] = _distanceBlue[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepRed[(int)actualcell.Item1 + 1, (int)actualcell.Item2] = (-1, 0);
                        list.Add(newcell);

                    }
                }
                z++;
            }
        }

        /// <summary>
        /// Calculate the path red units have to travel
        /// </summary>
        private void GetPathRed()
        {

            _distanceRed = new int[_table.Height, _table.Width];
            _stepBlue = new (int, int)[_table.Height, _table.Width];

            List<Tuple<int, int>> list = new List<Tuple<int, int>>();
            int z = 0;
            list.Clear();
            Tuple<int, int> v = new Tuple<int, int>(_table.RedCastle.Position.Y, _table.RedCastle.Position.X);

            for (int i = 0; i < _table.Height; i++)
            {
                for (int j = 0; j < _table.Width; j++)
                {
                    _distanceRed[i, j] = (int)(_table.Width * _table.Height + 1);
                    _stepBlue[i, j] = (0, 0);
                }
            }

            list.Add(v);
            _distanceRed[(int)v.Item1, (int)v.Item2] = 0;
            while (z < list.Count)
            {
                Tuple<int, int> actualcell = list[z];

                if (actualcell.Item1 - 1 > -1)
                {
                    if (_table[(int)actualcell.Item1 - 1, (int)actualcell.Item2] == null && _distanceRed[(int)actualcell.Item1 - 1, (int)actualcell.Item2] > _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 - 1, actualcell.Item2);
                        _distanceRed[(int)actualcell.Item1 - 1, (int)actualcell.Item2] = _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepBlue[(int)actualcell.Item1 - 1, (int)actualcell.Item2] = (1, 0);
                        list.Add(newcell);

                    }
                }
                if (actualcell.Item2 + 1 < _table.Width)
                {
                    if (_table[(int)actualcell.Item1, (int)actualcell.Item2 + 1] == null && _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2 + 1] > _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 + 1);
                        _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2 + 1] = _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepBlue[(int)actualcell.Item1, (int)actualcell.Item2 + 1] = (0, -1);
                        list.Add(newcell);

                    }
                }
                if (actualcell.Item2 - 1 > -1)
                {
                    if (_table[(int)actualcell.Item1, (int)actualcell.Item2 - 1] == null && _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2 - 1] > _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1, actualcell.Item2 - 1);
                        _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2 - 1] = _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepBlue[(int)actualcell.Item1, (int)actualcell.Item2 - 1] = (0, 1);
                        list.Add(newcell);

                    }
                }
                if (actualcell.Item1 + 1 < _table.Height)
                {
                    if (_table[(int)actualcell.Item1 + 1, (int)actualcell.Item2] == null && _distanceRed[(int)actualcell.Item1 + 1, (int)actualcell.Item2] > _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1)
                    {
                        Tuple<int, int> newcell = new Tuple<int, int>(actualcell.Item1 + 1, actualcell.Item2);
                        _distanceRed[(int)actualcell.Item1 + 1, (int)actualcell.Item2] = _distanceRed[(int)actualcell.Item1, (int)actualcell.Item2] + 1;
                        _stepBlue[(int)actualcell.Item1 + 1, (int)actualcell.Item2] = (-1, 0);
                        list.Add(newcell);

                    }
                }
                z++;
            }
        }

        /// <summary>
        /// Starts the simulation phase
        /// </summary>
        private void StartSimulation()
        {
            GetPathBlue();
            GetPathRed();
            if (Table.BlueUnitsCurrentPositions is null)
            {
                Table.BlueUnitsCurrentPositions = new List<List<List<Unit>>>();
            }
            if (Table.RedUnitsCurrentPositions is null)
            {
                Table.RedUnitsCurrentPositions = new List<List<List<Unit>>>();
            }
            InitializeSimulationMatrix();
        }

        /// <summary>
        /// Checks whether the game has ended
        /// </summary>
        private void CheckGameEnd()
        {
            if (_table.BlueCastle.Health <= 0)
            {
                if (_table.RedCastle.Health <= 0)
                {
                    _table.State = GameState.Menu;
                    OnGameOver(3); // döntetlen
                }
                else
                {
                    _table.State = GameState.Menu;
                    OnGameOver(2); //piros nyert
                }
            }
            else
            {
                if (_table.RedCastle.Health <= 0)
                {
                    _table.State = GameState.Menu;
                    OnGameOver(1); // kék nyert
                }
            }
        }
        
        /// <summary>
        /// Shhots a unit with the tower on the given cell
        /// </summary>
        /// <param name="y">Y coordinate of tower</param>
        /// <param name="x">X coordinate of tower</param>
        /// <param name="range">Range of tower</param>
        /// <param name="isBlue">Whether the tower is blue</param>
        /// <returns>The position of the unit that is shot or (-1, -1) if there are no units in range</returns>
        private Tuple<int,int> ShootUnit(int y, int x, double range, bool isBlue)
        {
            int y0 = (int)Math.Round(y - range);
            int x0 = (int)Math.Round(x - range);
            int y1 = (int)Math.Round(y + range);
            int x1 = (int)Math.Round(x + range);

            int minimum = _table.Width * _table.Height + 1;
            Tuple<int, int> unitpos = new Tuple<int,int>(-1,-1);

            for (int i = y0; i<= y1; i++)
            {
                for(int j = x0; j<=x1; j++)
                {
                   
                    if (i>=0 && i <_table.Height && j >=0 && j < _table.Width)
                    {
                        if (isBlue)
                        {
                            
                            if (Table.BlueUnitsCurrentPositions[i][j].Count != 0)
                            {
                                if (_distanceRed[i,j] < minimum)
                                {
                                    minimum = _distanceRed[i, j];
                                    unitpos = new Tuple<int,int>(i, j);
                                   
                                }
                            }
                        }
                        else
                        {
                            if (Table.RedUnitsCurrentPositions[i][j].Count != 0)
                            {
                                if (_distanceBlue[i, j] < minimum)
                                {
                                    minimum = _distanceBlue[i, j];
                                    unitpos = new Tuple<int, int>(i, j);
                                }
                            }
                        }
                    }
                }
            }
            
            return unitpos;
        }

        #endregion

        #region Private event methods

        /// <summary>
        /// This method runs after the gametable is created
        /// </summary>
        private void OnGameCreated()
        {
            if (GameCreated != null)
                GameCreated(this, new GameEventArgs(_table.BlueHealth, _table.BlueGold, _table.RedHealth, _table.RedGold));
        }

        /// <summary>
        /// This method runs when the game state changes in some way
        /// </summary>
        /// <param name="soldTower">Whether a tower was sold</param>
        private void OnGameAdvanced(bool soldTower = false)
        {
            if (GameAdvanced != null)
            {
                if (!soldTower)
                {
                    GameAdvanced(this, new GameEventArgs(_table.BlueHealth, _table.BlueGold, _table.RedHealth, _table.RedGold));
                }
                else
                {
                    GameAdvanced(this, new GameEventArgs(_table.BlueHealth, _table.BlueGold, _table.RedHealth, _table.RedGold)
                    {
                        SoldTower = true
                    });
                }
            }
        }

        /// <summary>
        /// This method runs when the game state changes in some way
        /// </summary>
        /// <param name="upgradedTower">Whether a tower was upgraded</param>
        /// <param name="y">Y coordinate of selected cell</param>
        /// <param name="x">X coordiante of selected cell</param>
        private void OnGameAdvanced(bool upgradedTower, int y, int x)
        {
            if (GameAdvanced != null)
            {
                if (!upgradedTower)
                {
                    GameAdvanced(this, new GameEventArgs(y, x));
                }
                else
                {
                    GameAdvanced(this, new GameEventArgs(y, x)
                    {
                        UpgradedTower = true
                    });
                }
            }
        }

        /// <summary>
        /// This method runs when the game state changes in some way
        /// </summary>
        /// <param name="whyIsNothingHappening">Message to be displayed</param>
        private void OnGameAdvanced(String whyIsNothingHappening)
        {
            if (GameAdvanced != null)
            {                
                GameAdvanced(this, new GameEventArgs(whyIsNothingHappening));
            }
        }

        /// <summary>
        /// This method runs when the game is over
        /// </summary>
        /// <param name="whoWon">Who won the game (1 - blue, 2 - red, 3 - draw)</param>
        private void OnGameOver(int whoWon)
        {
            if (GameOver != null)
                GameOver(this, new GameEventArgs(whoWon));
        }

        #endregion
    }
}
