﻿using System;
using System.Text;
using System.Linq;
using System.Collections.Generic;
using System.IO;
using System.Threading.Tasks;
using System.Diagnostics;

namespace ScaraBattle.Persistence
{
    /// <summary>
    /// Save and load game state using files
    /// </summary>
    public class GameFileDataAccess : IGameDataAccess
    {
        private async Task<GameTable> LoadTable(StreamReader reader)
        {
            int[] size = (await reader.ReadLineAsync()).Split(' ').Select(n => Int32.Parse(n)).ToArray();
            String st = await reader.ReadLineAsync();
            GameState state = st == "blue" ? GameState.Blue : (st == "red" ? GameState.Red : GameState.MapEditor);

            return new GameTable(size[0], size[1], state);
        }

        private async Task LoadMap(StreamReader reader, GameTable table)
        {
            int r = 0;
            int c = 0;
            for (int i = 0; i < table.Height; i++)
            {
                String[] row = (await reader.ReadLineAsync()).Split(' ');
                foreach (String obj in row)
                {
                    switch (obj)
                    {
                        case "D":
                            table[r, c] = Dune.Instance;
                            break;
                        case "O":
                            table[r, c] = Oasis.Instance;
                            break;
                        case "R":
                            table[r, c] = Ruins.Instance;
                            break;
                        default:
                            table[r, c] = null;
                            break;
                    }
                    c++;
                }
                r++;
                c = 0;
            }
        }

        private async Task LoadCastles(StreamReader reader, GameTable table)
        {
            int castleNum = Int32.Parse(await reader.ReadLineAsync());
            if (castleNum == 1)
            {
                int[] blueCastle;
                blueCastle = (await reader.ReadLineAsync()).Split(' ').Select(n => Int32.Parse(n)).ToArray();
                table.BlueCastle = new Castle() { IsBlue = true };
                table[blueCastle[0], blueCastle[1]] = table.BlueCastle;
                table.BlueCastle.Position = (blueCastle[0], blueCastle[1]);
            }
            castleNum = Int32.Parse(await reader.ReadLineAsync());
            if (castleNum == 1)
            {
                int[] redCastle;
                redCastle = (await reader.ReadLineAsync()).Split(' ').Select(n => Int32.Parse(n)).ToArray();
                table.RedCastle = new Castle() { IsBlue = false };
                table[redCastle[0], redCastle[1]] = table.RedCastle;
                table.RedCastle.Position = (redCastle[0], redCastle[1]);
            }
        }

        private async Task LoadBarracks(StreamReader reader, GameTable table)
        {
            int blueBarrack = Int32.Parse(await reader.ReadLineAsync());
            if (blueBarrack != 0)
            {
                table.BlueBarracks = new (int, int)[blueBarrack];
                int indexBlue = 0;
                int[] coords = (await reader.ReadLineAsync()).Split(' ').Select(e => Int32.Parse(e)).ToArray();
                for (int i = 0; i < coords.Length; i += 2)
                {
                    table.BlueBarracks[indexBlue] = (coords[i], coords[i + 1]);
                    table[coords[i], coords[i + 1]] = new Barrack() { IsBlue = true };
                    indexBlue++;
                }
            }

            int redBarrack = Int32.Parse(await reader.ReadLineAsync());
            if (redBarrack != 0)
            {
                table.RedBarracks = new (int, int)[redBarrack];
                int indexRed = 0;
                int[] coords = (await reader.ReadLineAsync()).Split(' ').Select(e => Int32.Parse(e)).ToArray();
                for (int i = 0; i < coords.Length; i += 2)
                {
                    table.RedBarracks[indexRed] = (coords[i], coords[i + 1]);
                    table[coords[i], coords[i + 1]] = new Barrack() { IsBlue = false };
                    indexRed++;
                }
            }
        }

        private async Task LoadTowers(StreamReader reader, GameTable table)
        {
            int blueTowers = Int32.Parse(await reader.ReadLineAsync());
            table.BlueTowerPositions = new List<(int, int)>();
            if (blueTowers != 0)
            {
                for (int i = 0; i < blueTowers; i++)
                {
                    float[] row = (await reader.ReadLineAsync()).Split(' ').Select(e => float.Parse(e)).ToArray();

                    table.BlueTowerPositions.Add(((int)row[0], (int)row[1]));

                    Tower t;
                    if (row[4] < 3)
                    {
                        t = new SmallTower();
                    }
                    else if (row[4] < 5)
                    {
                        t = new MediumTower();
                    }
                    else
                    {
                        t = new LargeTower();
                    }
                    t.IsBlue = true;
                    t.Range = row[2];
                    t.Damage = (int)row[3];
                    t.Level = (int)row[4];
                    table[(int)row[0], (int)row[1]] = t;
                }
            }


            int redTowers = Int32.Parse(await reader.ReadLineAsync());
            table.RedTowerPositions = new List<(int, int)>();
            if (redTowers != 0)
            {
                for (int i = 0; i < redTowers; i++)
                {
                    float[] row = (await reader.ReadLineAsync()).Split(' ').Select(e => float.Parse(e)).ToArray();
                    table.RedTowerPositions.Add(((int)row[0], (int)row[1]));
                    Tower t;
                    if (row[4] < 3)
                    {
                        t = new SmallTower();
                    }
                    else if (row[4] < 5)
                    {
                        t = new MediumTower();
                    }
                    else
                    {
                        t = new LargeTower();
                    }
                    t.IsBlue = false;
                    t.Range = row[2];
                    t.Damage = (int)row[3];
                    t.Level = (int)row[4];
                    table[(int)row[0], (int)row[1]] = t;
                }
            }
        }

        /// <summary>
        /// Load game state from file
        /// </summary>
        /// <param name="path">Path of the file</param>
        /// <returns>Game state parameters</returns>
        public async Task<GameTable> LoadAsync(String path)
        {
            try
            {
                using (StreamReader reader = new StreamReader(path))
                {
                    GameTable table = await LoadTable(reader);
                    await LoadMap(reader, table);
                    await LoadCastles(reader, table);
                    await LoadBarracks(reader, table);

                    table.BlueGold = Int32.Parse(await reader.ReadLineAsync());
                    table.BlueHealth = Int32.Parse(await reader.ReadLineAsync());

                    table.RedGold = Int32.Parse(await reader.ReadLineAsync());
                    table.RedHealth = Int32.Parse(await reader.ReadLineAsync());

                    await LoadTowers(reader, table);

                    table.BlueUnits = new List<Unit>();
                    table.BlueUnitsCurrentPositions = new List<List<List<Unit>>>();

                    table.RedUnits = new List<Unit>();
                    table.RedUnitsCurrentPositions = new List<List<List<Unit>>>();

                    return table;
                }
            }
            catch
            {
                throw new GameDataException();
            }
        }

        private async Task SaveTable(StreamWriter writer, GameTable table)
        {
            await writer.WriteLineAsync(table.Height.ToString() + " " + table.Width.ToString());
            await writer.WriteLineAsync(table.State == GameState.Red ? "red" : (table.State == GameState.Blue ? "blue" : "mapeditor"));
        }

        private async Task SaveMap(StreamWriter writer, GameTable table)
        {
            StringBuilder sb = new StringBuilder("", table.Width);
            for (int i = 0; i < table.Height; i++)
            {
                for (int j = 0; j < table.Width; j++)
                {
                    if (table[i, j] is Dune)
                    {
                        sb.Append('D');
                    }
                    else if (table[i, j] is Oasis)
                    {
                        sb.Append('O');
                    }
                    else if (table[i, j] is Ruins)
                    {
                        sb.Append('R');
                    }
                    else
                    {
                        sb.Append('E');
                    }

                    if (j != table.Width - 1)
                    {
                        sb.Append(' ');
                    }
                }
                await writer.WriteLineAsync(sb.ToString());
                sb.Clear();
            }
        }

        private async Task SaveCastles(StreamWriter writer, GameTable table)
        {
            if (table.BlueCastle is null)
            {
                await writer.WriteLineAsync("0");
            }
            else
            {
                await writer.WriteLineAsync("1");
                await writer.WriteLineAsync(table.BlueCastle.Position.Y.ToString() + " " + table.BlueCastle.Position.X.ToString());
            }
            if (table.RedCastle is null)
            {
                await writer.WriteLineAsync("0");
            }
            else
            {
                await writer.WriteLineAsync("1");
                await writer.WriteLineAsync(table.RedCastle.Position.Y.ToString() + " " + table.RedCastle.Position.X.ToString());
            }
        }

        private async Task SaveBarracks(StreamWriter writer, GameTable table)
        {
            if (table.BlueBarracks is null || table.BlueBarracks.Length == 0)
            {
                await writer.WriteLineAsync("0");
            }
            else
            {
                await writer.WriteLineAsync(table.BlueBarracks.Length.ToString());
                await writer.WriteLineAsync(String.Join(" ", table.BlueBarracks.Select((coordinates) => coordinates.Item1.ToString() + " " + coordinates.Item2.ToString()).ToArray()));
            }
            if (table.RedBarracks is null || table.RedBarracks.Length == 0)
            {
                await writer.WriteLineAsync("0");
            }
            else
            {
                await writer.WriteLineAsync(table.RedBarracks.Length.ToString());
                await writer.WriteLineAsync(String.Join(" ", table.RedBarracks.Select((coordinates) => coordinates.Item1.ToString() + " " + coordinates.Item2.ToString()).ToArray()));
            }
        }

        private async Task SaveTowers(StreamWriter writer, GameTable table)
        {
            if (table.BlueTowerPositions is null || table.BlueTowerPositions.Count == 0)
            {
                await writer.WriteLineAsync("0");
            }
            else
            {
                await writer.WriteLineAsync(table.BlueTowerPositions.Count.ToString());
                foreach ((int Y, int X) pos in table.BlueTowerPositions)
                {
                    Tower t = (Tower)table[pos.Y, pos.X];
                    float[] row = { pos.Y, pos.X, t.Range, t.Damage, t.Level };
                    await writer.WriteLineAsync(String.Join(" ", row.Select(e => e.ToString())));
                }
            }
            if (table.RedTowerPositions is null || table.RedTowerPositions.Count == 0)
            {
                await writer.WriteLineAsync("0");
            }
            else
            {
                await writer.WriteLineAsync(table.RedTowerPositions.Count.ToString());
                foreach ((int Y, int X) pos in table.RedTowerPositions)
                {
                    Tower t = (Tower)table[pos.Y, pos.X];
                    float[] row = { pos.Y, pos.X, t.Range, t.Damage, t.Level };

                    await writer.WriteLineAsync(String.Join(" ", row.Select(e => e.ToString())));
                }
            }
        }

        /// <summary>
        /// Save game state to a file
        /// </summary>
        /// <param name="path">Path of the file</param>
        /// <param name="table">Current state of the game</param>
        public async Task SaveAsync(String path, GameTable table)
        {
            try
            {
                using (StreamWriter writer = new StreamWriter(path))
                {
                    await SaveTable(writer, table);
                    await SaveMap(writer, table);
                    await SaveCastles(writer, table);
                    await SaveBarracks(writer, table);

                    await writer.WriteLineAsync(table.BlueGold.ToString());
                    await writer.WriteLineAsync(table.BlueHealth.ToString());

                    await writer.WriteLineAsync(table.RedGold.ToString());
                    await writer.WriteLineAsync(table.RedHealth.ToString());

                    await SaveTowers(writer, table);
                }
            }
            catch
            {
                throw new GameDataException();
            }
        }
    }
}
