﻿using System;
using System.Collections.ObjectModel;
using ScaraBattle.Model;
using ScaraBattle.Persistence;
using System.Windows.Threading;
using System.Collections.Generic;
using System.Diagnostics;

namespace ScaraBattle.ViewModel
{
    /// <summary>
    /// Enum for the type of information that needs to be shown on the screen
    /// </summary>
    public enum InfoType
    {
        Tower,
        General,
        Unit,
        Nothing,
        Text
    }

    /// <summary>
    /// The view model class of the game
    /// </summary>
    class GameViewModel : ViewModelBase
    {
        #region Fields

        private GameModel _model;

        // nothing
        private const String NothingImage = "NothingImage";
        // obstacles
        private const String DuneImage = "DuneImage";
        private const String OasisImage = "OasisImage";
        private const String RuinsImage = "RuinsImage";
        // castle, barrack
        private const String CastleImage = "CastleImage";
        private const String BarrackImage = "BarrackImage";

        private const String CastleImageBlue = "CastleImageBlue";
        private const String BarrackImageBlue = "BarrackImageBlue";

        private const String CastleImageRed = "CastleImageRed";
        private const String BarrackImageRed = "BarrackImageRed";

        // units
        private const String ScarabImageSmall = "ScarabImageSmall";
        private const String ScarabImage = "ScarabImage";
        private const String ScarabImageBlueSmall = "ScarabImageBlueSmall";
        private const String ScarabImageRedSmall = "ScarabImageRedSmall";
        private const String ScarabImageBlue = "ScarabImageBlue";
        private const String ScarabImageRed = "ScarabImageRed";
        private const String ScarabImageMore = "ScarabImageMore";
        private const String ScarabImageDead = "ScarabImageDead";
        private const String ScarabImageGhost = "ScarabImageGhost";

        // towers
        private const String SmallTowerImage = "SmallTowerImage";
        private const String MediumTowerImage = "MediumTowerImage";
        private const String LargeTowerImage = "LargeTowerImage";

        private const String SmallTowerImageBlue = "SmallTowerImageBlue";
        private const String MediumTowerImageBlue = "MediumTowerImageBlue";
        private const String LargeTowerImageBlue = "LargeTowerImageBlue";

        private const String SmallTowerImageRed = "SmallTowerImageRed";
        private const String MediumTowerImageRed = "MediumTowerImageRed";
        private const String LargeTowerImageRed = "LargeTowerImageRed";

        private int _offsetY;
        private int _offsetX;

        private SelectedObject _selectedObject = SelectedObject.Nothing;

        private (int Y, int X) _selectedCell;
        private (int Y, int X) _selectedTower;

        private DispatcherTimer _timer;

        #endregion

        #region Properties

        /// <summary>
        /// New game delegate command
        /// </summary>
        public DelegateCommand NewGameCommand { get; private set; }

        /// <summary>
        /// Map creation delegate command
        /// </summary>
        public DelegateCommand CreateMapCommand { get; private set; }

        /// <summary>
        /// Load game delegate command
        /// </summary>
        public DelegateCommand LoadGameCommand { get; private set; }

        /// <summary>
        /// Save game delegate command
        /// </summary>
        public DelegateCommand SaveGameCommand { get; private set; }

        /// <summary>
        /// Exit game delegate command
        /// </summary>
        public DelegateCommand ExitGameCommand { get; private set; }

        /// <summary>
        /// Show help delegate command
        /// </summary>
        public DelegateCommand HelpCommand { get; private set; }

        /// <summary>
        /// Show help delegate command
        /// </summary>
        public DelegateCommand HelpTowerCommand { get; private set; }

        /// <summary>
        /// Show helpunit delegate command
        /// </summary>
        public DelegateCommand HelpUnitCommand { get; private set; }

        /// <summary>
        /// Show helpmapeditor delegate command
        /// </summary>
        public DelegateCommand HelpMapeditorCommand { get; private set; }

        /// <summary>
        /// Open map editor delegate command
        /// </summary>
        public DelegateCommand MapEditorCommand { get; private set; }

        /// <summary>
        /// Move camera up delegate command
        /// </summary>
        public DelegateCommand CameraUpCommand { get; private set; }

        /// <summary>
        /// Move camera left delegate command
        /// </summary>
        public DelegateCommand CameraLeftCommand { get; private set; }

        /// <summary>
        /// Move camera right delegate command
        /// </summary>
        public DelegateCommand CameraRightCommand { get; private set; }

        /// <summary>
        /// Move camera down delegate command
        /// </summary>
        public DelegateCommand CameraDownCommand { get; private set; }

        /// <summary>
        /// Sell tower delegate command
        /// </summary>
        public DelegateCommand SellTowerCommand { get; private set; }

        /// <summary>
        /// Upgrade tower delegate command
        /// </summary>
        public DelegateCommand UpgradeTowerCommand { get; private set; }

        /// <summary>
        /// Step game state delegate command
        /// </summary>
        public DelegateCommand StepGameStateCommand { get; private set; }

        /// <summary>
        /// Collection of game fields
        /// </summary>
        public ObservableCollection<GameField> Fields { get; set; }

        /// <summary>
        /// Collection of menu items in the game
        /// </summary>
        public ObservableCollection<MenuItem> GameMenuItems { get; set; }

        /// <summary>
        /// Collection of menu items in the map editor
        /// </summary>
        public ObservableCollection<MenuItem> MapEditorMenuItems { get; set; }

        /// <summary>
        /// Collection of unit list items during simulation
        /// </summary>
        public ObservableCollection<UnitListItem> UnitListItems { get; set; }

        /// <summary>
        /// Height of grid
        /// </summary>
        public int GridHeight { get; set; }

        /// <summary>
        /// Width of grid
        /// </summary>
        public int GridWidth { get; set; }

        public string SelectedItem { get; set; }

        public int BlueGold { get { return _model.Table == null ? 0 : _model.Table.BlueGold; } }

        public int RedGold { get { return _model.Table == null ? 0 : _model.Table.RedGold; } }
        public int BlueHealth { get { return _model.Table == null ? 0 : _model.Table.BlueHealth; } }
        public int RedHealth { get { return _model.Table == null ? 0 : _model.Table.RedHealth; } }
        public int BlueSmallUnits { get { return _model.Table == null ? 0 : _model.Table.BlueSmallUnits; } }
        public int BlueLargeUnits { get { return _model.Table == null ? 0 : _model.Table.BlueLargeUnits; } }
        public int RedSmallUnits { get { return _model.Table == null ? 0 : _model.Table.RedSmallUnits; } }
        public int RedLargeUnits { get { return _model.Table == null ? 0 : _model.Table.RedLargeUnits; } }

        public int TowerDamage { get; set; }
        public float TowerRange { get; set; }
        public int TowerLevel { get; set; }
        public int TowerSellCost { get; set; }
        public int TowerUpgradeCost { get; set; }

        public string InfoName { get; set; }

        public string InfoImage { get; set; }

        public string WhyIsNothingHappening { get; set; }

        public InfoType VisibleInfo { get; set; }

        public string StepGameButtonText { get; set; }

        private bool canGoUp;
        public bool CanGoUp
        {
            get { return canGoUp; }
            set { canGoUp = value;
                OnPropertyChanged(nameof(CanGoUp));
            }
        }
        private bool canGoDown;
        public bool CanGoDown
        {
            get { return canGoDown; }
            set
            {
                canGoDown = value;
                OnPropertyChanged(nameof(CanGoDown));
            }
        }
        private bool canGoLeft;
        public bool CanGoLeft
        {
            get { return canGoLeft; }
            set
            {
                canGoLeft = value;
                OnPropertyChanged(nameof(CanGoLeft));
            }
        }
        private bool canGoRight;
        public bool CanGoRight
        {
            get { return canGoRight; }
            set
            {
                canGoRight = value;
                OnPropertyChanged(nameof(CanGoRight));
            }
        }

        public GameState State { get { return _model.Table == null ? GameState.Menu : _model.Table.State; } }

        #endregion

        #region Events

        /// <summary>
        /// New game event
        /// </summary>
        public event EventHandler NewGame;

        /// <summary>
        /// Map creation event
        /// </summary>
        public event EventHandler CreateMap;

        /// <summary>
        /// Load game event
        /// </summary>
        public event EventHandler LoadGame;

        /// <summary>
        /// Save game event
        /// </summary>
        public event EventHandler SaveGame;

        /// <summary>
        /// Exit game event
        /// </summary>
        public event EventHandler ExitGame;

        /// <summary>
        /// Show help event
        /// </summary>
        public event EventHandler Help;

        /// <summary>
        /// Show help event to towers
        /// </summary>
        public event EventHandler HelpTower;

        /// <summary>
        /// Show help event to units
        /// </summary>
        public event EventHandler HelpUnit;

        /// <summary>
        /// Show help event to map editor
        /// </summary>
        public event EventHandler HelpMapeditor;

        /// <summary>
        /// Open map editor event
        /// </summary>
        public event EventHandler EditMap;

        /// <summary>
        /// This method handles the passing of time
        /// </summary>
        /// <param name="sender">Sender object</param>
        /// <param name="e">EventArgs object</param>
        private void TimerTick(object sender, EventArgs e)
        {
            if (_model.Table.State == GameState.Simulation)
            {
                _model.Simulate();
            }
            
        }


        #endregion

        #region Constructors

        /// <summary>
        /// Constructor for the viewmodel
        /// </summary>
        /// <param name="model">The model object for the viewmodel</param>
        public GameViewModel(GameModel model)
        {
            // connect model to viewmodel
            _model = model;
            _model.GameAdvanced += new EventHandler<GameEventArgs>(Model_GameAdvanced);
            _model.GameCreated += new EventHandler<GameEventArgs>(Model_GameCreated);
            _model.GameOver += new EventHandler<GameEventArgs>((obj, e) => RefreshTable());

            _timer = new DispatcherTimer();
            _timer.Tick += TimerTick;
            _timer.Interval = new TimeSpan(0, 0, 1);
            _timer.Start();

            _selectedTower = (-1, -1);


            // handle commands
            NewGameCommand = new DelegateCommand(param => OnNewGame());
            LoadGameCommand = new DelegateCommand(param => OnLoadGame());
            SaveGameCommand = new DelegateCommand(param => OnSaveGame());
            ExitGameCommand = new DelegateCommand(param => OnExitGame());
            HelpCommand = new DelegateCommand(param => OnHelpRequested());
            HelpTowerCommand = new DelegateCommand(param => OnHelpRequestedTower());
            HelpUnitCommand = new DelegateCommand(param => OnHelpRequestedUnit());
            HelpMapeditorCommand = new DelegateCommand(param => OnHelpRequestedMapeditor());
            MapEditorCommand = new DelegateCommand(param => OnEditMap());

            CameraUpCommand = new DelegateCommand(param => OnCameraUp());
            CameraLeftCommand = new DelegateCommand(param => OnCameraLeft());
            CameraRightCommand = new DelegateCommand(param => OnCameraRight());
            CameraDownCommand = new DelegateCommand(param => OnCameraDown());

            InitializeMenuItems();

            CreateMapCommand = new DelegateCommand(param => OnCreateMap());

            SellTowerCommand = new DelegateCommand(param => OnSellTower());
            UpgradeTowerCommand = new DelegateCommand(param => OnUpgradeTower());

            StepGameStateCommand = new DelegateCommand(param => OnStepGameState());
        }

        #endregion

        #region Public methods

        /// <summary>
        /// Initialize the game fields that can be seen on the window
        /// </summary>
        public void GenerateTiles()
        {
            GridHeight = 10;
            GridWidth = 10;

            _offsetX = (_model.Table.Width / 2) - (GridWidth/2);
            _offsetY = (_model.Table.Height / 2) - (GridHeight/2);
            CanGoUp = _offsetY > 0;
            CanGoDown = _offsetY < _model.Table.Height - GridHeight;
            CanGoLeft = _offsetX > 0;
            CanGoRight = _offsetX < _model.Table.Width - GridWidth;

            Fields = new ObservableCollection<GameField>();
            for (int i = 0; i < GridHeight; i++)
            {
                for (int j = 0; j < GridWidth; j++)
                {
                    Fields.Add(new GameField
                    {
                        X = i,
                        Y = j,
                        Image = "",
                        Color = _model.Table.ColorStandard,
                        CellCommand = new DelegateCommand(param => SelectCell((((int Y, int X))param).Y, (((int Y, int X))param).X))
                    });
                }
            }

            RefreshTable();
        }

        /// <summary>
        /// Refresh each field of the gametable
        /// </summary>
        public void RefreshTable()
        {
            foreach (GameField field in Fields)
            {
                field.Image = "";
                field.Color = _model.Table.ColorStandard;
            }

            for (int i = 0; i < GridHeight; ++i)
            {
                for (int j = 0; j < GridWidth; ++j)
                {
                    object cell = _model.Table[i + _offsetY, j + _offsetX];

                    if (cell is Dune)
                    {
                        GetField(i, j).Image = DuneImage;
                    }
                    else if (cell is Oasis)
                    {
                        GetField(i, j).Image = OasisImage;
                    }
                    else if (cell is Ruins)
                    {
                        GetField(i, j).Image = RuinsImage;
                    }
                    else if (cell is Castle c)
                    {
                        if (c.IsBlue)
                        {
                          GetField(i, j).Image = CastleImageBlue;
                        }
                        else {
                            GetField(i, j).Image = CastleImageRed;
                        }
                    }
                    else if (cell is Barrack b)
                    {
                        if (b.IsBlue)
                        {
                            GetField(i, j).Image = BarrackImageBlue;
                        }
                        else {
                            GetField(i, j).Image = BarrackImageRed;
                        }
                    }
                }
            }

            if (_model.Table.State != GameState.MapEditor)
            {
                if (_selectedTower != (-1,-1) && (_selectedObject == SelectedObject.Nothing ||
                    _selectedObject == SelectedObject.LargeTower || _selectedObject == SelectedObject.MediumTower ||
                    _selectedObject == SelectedObject.SmallTower))
                {
                    foreach (Tuple<int, int> coord in _model.GetTowerRange(_selectedTower.Y + _offsetY, _selectedTower.X +_offsetX)){
                        int x = coord.Item2;
                        int y = coord.Item1;
                        if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                            0 <= y - _offsetY && y - _offsetY < GridHeight)
                        {
                            GetField(y - _offsetY, x - _offsetX).Color = _model.Table.ColorMarked;
                        }
                        // TODO
                    }
                }
                // Move units
                if (_model.Table.State == GameState.Simulation)
                {
                    // Blue units
                    for (int i = 0; i<_model.Table.BlueUnitsCurrentNum; i++)
                    {
                        int y = _model.Table.BlueUnits[i].Y;
                        int x = _model.Table.BlueUnits[i].X;
                        if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                            0 <= y - _offsetY && y - _offsetY < GridHeight)
                        {
                            if (_model.IsTwoUnits(y,x))
                            {
                                GetField(y - _offsetY, x - _offsetX).Image = ScarabImageMore;
                            }
                            else if (_model.Table.BlueUnits[i] is LargeUnit)
                            {
                                GetField(y - _offsetY, x - _offsetX).Image = ScarabImageBlue;
                            }
                            else
                            {
                                GetField(y - _offsetY, x - _offsetX).Image = ScarabImageBlueSmall;
                            }
                        }
                    }

                    // Red units
                    for (int i = 0; i < _model.Table.RedUnitsCurrentNum; i++)
                    {
                        int y = _model.Table.RedUnits[i].Y;
                        int x = _model.Table.RedUnits[i].X;
                        if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                            0 <= y - _offsetY && y - _offsetY < GridHeight)
                        {
                            if (_model.IsTwoUnits(y, x))
                            {
                                GetField(y - _offsetY, x - _offsetX).Image = ScarabImageMore;
                            }
                            else if (_model.Table.RedUnits[i] is LargeUnit)
                            {
                                GetField(y - _offsetY, x - _offsetX).Image = ScarabImageRed;
                            }
                            else
                            {
                                GetField(y - _offsetY, x - _offsetX).Image = ScarabImageRedSmall;
                            }
                        }
                    }

                    // Background color of injured units
                    foreach (Tuple<int, int> pos in _model.GetUnitsShot())
                    {
                        int x = pos.Item2;
                        int y = pos.Item1;
                        if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                            0 <= y - _offsetY && y - _offsetY < GridHeight)
                        {
                            GetField(y - _offsetY, x - _offsetX).Color = _model.Table.ColorInjured;
                        }
                    }

                    // Background color of killed units
                    foreach (Tuple<int, int> pos in _model.GetUnitsKilled())
                    {
                        int x = pos.Item2;
                        int y = pos.Item1;
                        if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                            0 <= y - _offsetY && y - _offsetY < GridHeight)
                        {
                            // GetField(y - _offsetY, x - _offsetX).Color = _model.Table.ColorInjured;
                            GetField(y - _offsetY, x - _offsetX).Image = ScarabImageGhost;
                        }
                    }


                }

                foreach ((int Y, int X) t in _model.Table.BlueTowerPositions)
                {
                    int y = t.Y;
                    int x = t.X;

                    if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                        0 <= y - _offsetY && y - _offsetY < GridHeight)
                    {
                        object cell = _model.Table[y, x];

                        if (cell is SmallTower)
                        {
                            GetField(y - _offsetY, x - _offsetX).Image = SmallTowerImageBlue;
                        }
                        else if (cell is MediumTower)
                        {
                            GetField(y - _offsetY, x - _offsetX).Image = MediumTowerImageBlue;
                        }
                        else if (cell is LargeTower)
                        {
                            GetField(y - _offsetY, x - _offsetX).Image = LargeTowerImageBlue;
                        }
                    }
                }

                foreach ((int Y, int X) t in _model.Table.RedTowerPositions)
                {
                    int y = t.Y;
                    int x = t.X;

                    if (0 <= x - _offsetX && x - _offsetX < GridWidth &&
                        0 <= y - _offsetY && y - _offsetY < GridHeight)
                    {
                        object cell = _model.Table[y, x];

                        if (cell is SmallTower)
                        {
                            GetField(y - _offsetY, x - _offsetX).Image = SmallTowerImageRed;
                        }
                        else if (cell is MediumTower)
                        {
                            GetField(y - _offsetY, x - _offsetX).Image = MediumTowerImageRed;
                        }
                        else if (cell is LargeTower)
                        {
                            GetField(y - _offsetY, x - _offsetX).Image = LargeTowerImageRed;
                        }
                    }
                }
            }

            OnPropertyChanged(nameof(Fields));
            OnPropertyChanged(nameof(SelectedItem));
            OnPropertyChanged(nameof(BlueGold));
            OnPropertyChanged(nameof(RedGold));
            OnPropertyChanged(nameof(BlueHealth));
            OnPropertyChanged(nameof(RedHealth));
            OnPropertyChanged(nameof(BlueSmallUnits));
            OnPropertyChanged(nameof(BlueLargeUnits));
            OnPropertyChanged(nameof(RedSmallUnits));
            OnPropertyChanged(nameof(RedLargeUnits));
            OnPropertyChanged(nameof(State));
        }

        #endregion

        #region Private methods

        /// <summary>
        /// Initialization of the menu items that can be seen on the window
        /// </summary>
        private void InitializeMenuItems()
        {
            InitializeGameMenuItems();
            InitializeMapEditorMenuItems();
        }

        /// <summary>
        /// Initialization of game menu items (towers and units)
        /// </summary>
        private void InitializeGameMenuItems()
        {
            GameMenuItems = new ObservableCollection<MenuItem>();

            // Small tower
            GameMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.SmallTower,
                Price = SmallTower.Price,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param)),
                Damage = new SmallTower().Damage,
                Range = Math.Round(new SmallTower().Range, 2)
            });

            // Medium tower
            GameMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.MediumTower,
                Price = MediumTower.Price,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param)),
                Damage = new MediumTower().Damage,
                Range = Math.Round(new MediumTower().Range, 2)
            });

            // Large tower
            GameMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.LargeTower,
                Price = LargeTower.Price,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param)),
                Damage = new LargeTower().Damage,
                Range = Math.Round(new LargeTower().Range, 2)
            });

            // Small unit
            GameMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.SmallUnit,
                Price = SmallUnit.Price,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });

            // Large unit
            GameMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.LargeUnit,
                Price = LargeUnit.Price,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });
        }


        /// <summary>
        /// Initialization of map editor menu items (caslte, barrack and obstacles)
        /// </summary>
        private void InitializeMapEditorMenuItems()
        {
            MapEditorMenuItems = new ObservableCollection<MenuItem>();

            // Dune
            MapEditorMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.Dune,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });

            // Oasis
            MapEditorMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.Oasis,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });

            // Ruins
            MapEditorMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.Ruins,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });

            // Castle
            MapEditorMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.Castle,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });

            // Barrack
            MapEditorMenuItems.Add(new MenuItem()
            {
                Type = SelectedObject.Barrack,
                SelectCommand = new DelegateCommand(param => OnSelectMenuItem((SelectedObject)param))
            });
        }

        /// <summary>
        /// Get a field from the view
        /// </summary>
        /// <param name="i">Y coordinate of field</param>
        /// <param name="j">X coordinate of field</param>
        /// <returns>The field with the given coordinates</returns>
        private GameField GetField(int i, int j)
        {
            return Fields[i * GridWidth + j];
        }

        /// <summary>
        /// Given the coordinates of a tower, this method shows all information about that tower on the screen. The method needs to be called without camera offsets
        /// </summary>
        /// <param name="y">Y coordinate of tower</param>
        /// <param name="x">X coordinate of tower</param>
        private void ShowSelectedTower(int y, int x)
        {
            object cell = _model.Table[y + _offsetY, x + _offsetX];

            // show information on the screen
            if (cell is Tower)
            {
                _selectedTower = (y, x);
                _selectedCell = (y + _offsetY, x + _offsetX);

                TowerDamage = ((Tower)cell).Damage;
                float ize = ((Tower)cell).Range;
                TowerRange = (float)Math.Round(ize, 2);
                TowerLevel = ((Tower)cell).Level;
                TowerSellCost = ((Tower)cell).SellCost;
                TowerUpgradeCost = ((Tower)cell).UpgradeCost;
                VisibleInfo = InfoType.Tower;

                if (cell is SmallTower t)
                {
                    InfoName = "Kis torony";
                    InfoImage = t.IsBlue ? SmallTowerImageBlue : SmallTowerImageRed;
                }
                else if (cell is MediumTower m)
                {
                    InfoName = "Közepes torony";
                    InfoImage = m.IsBlue ? MediumTowerImageBlue : MediumTowerImageRed;
                }
                else if (cell is LargeTower l)
                {
                    InfoName = "Nagy torony";
                    InfoImage = l.IsBlue ? LargeTowerImageBlue : LargeTowerImageRed;
                }

                OnPropertyChanged(nameof(TowerDamage));
                OnPropertyChanged(nameof(TowerRange));
                OnPropertyChanged(nameof(TowerLevel));
                OnPropertyChanged(nameof(TowerSellCost));
                OnPropertyChanged(nameof(TowerUpgradeCost));
                OnPropertyChanged(nameof(VisibleInfo));
                OnPropertyChanged(nameof(InfoName));
                OnPropertyChanged(nameof(InfoImage));
            }
        }
        
        /// <summary>
        /// Select a game field on the screen
        /// </summary>
        /// <param name="y">Y coordinate of field</param>
        /// <param name="x">X coordinate of field</param>
        private void SelectCell(int y, int x)
        {
            _selectedTower = (-1, -1);
            if (_model.Table.State == GameState.MapEditor)
            {
                _model.PlaceObject(y + _offsetY, x + _offsetX, _selectedObject);
            }
            else if (_model.Table.State == GameState.Simulation)
            {
                List<Unit> units = new List<Unit>();

                foreach (Unit u in _model.Table.BlueUnits)
                {
                    if (u.Y == y + _offsetY && u.X == x + _offsetX)
                    {
                        units.Add(u);
                    }
                }

                foreach (Unit u in _model.Table.RedUnits)
                {
                    if (u.Y == y + _offsetY && u.X == x + _offsetX)
                    {
                        units.Add(u);
                    }
                }

                if (units.Count > 0)
                {
                    VisibleInfo = InfoType.Unit;
                    UnitListItems = new ObservableCollection<UnitListItem>();

                    foreach (Unit u in units)
                    {
                        SelectedObject unitType = SelectedObject.Nothing;
                        if (u is SmallUnit)
                        {
                            unitType = SelectedObject.SmallUnit;
                        }
                        else if (u is LargeUnit)
                        {
                            unitType = SelectedObject.LargeUnit;
                        }

                        UnitListItems.Add(new UnitListItem() { Health = u.Health, Type = unitType, IsBlue = u.IsBlue });
                    }

                    OnPropertyChanged(nameof(VisibleInfo));
                    OnPropertyChanged(nameof(UnitListItems));
                }
                else
                {
                    VisibleInfo = InfoType.Nothing;
                    OnPropertyChanged(nameof(VisibleInfo));
                    return;
                }
            }
            else if (_selectedObject != SelectedObject.Nothing) // place an object (e.g. a tower)
            {
                _model.PlaceObject(y + _offsetY, x + _offsetX, _selectedObject);
                ShowSelectedTower(y, x); // after placing the tower, all of its information is shown
            }
            else
            {
                object cell = _model.Table[y + _offsetY, x + _offsetX];

                // show information on screen
                if (cell is Tower)
                {
                    ShowSelectedTower(y, x);
                }

                else if (cell is Obstacle || cell is Castle || cell is Barrack)
                {
                    _selectedCell = (y + _offsetY, x + _offsetX);

                    VisibleInfo = InfoType.General;

                    if (cell is Dune)
                    {
                        InfoName = "Dűne";
                        InfoImage = "DuneImage";
                    }
                    if (cell is Oasis)
                    {
                        InfoName = "Oázis";
                        InfoImage = "OasisImage";
                    }
                    if (cell is Ruins)
                    {
                        InfoName = "Rom";
                        InfoImage = "RuinsImage";
                    }
                    if (cell is Castle c)
                    {
                        InfoName = "Kastély";
                        InfoImage = c.IsBlue ? CastleImageBlue : CastleImageRed;
                    }
                    if (cell is Barrack b)
                    {
                        InfoName = "Barakk";
                        InfoImage = b.IsBlue ? BarrackImageBlue : BarrackImageRed;
                    };
                    OnPropertyChanged(nameof(VisibleInfo));
                    OnPropertyChanged(nameof(InfoName));
                    OnPropertyChanged(nameof(InfoImage));
                }
                else
                {
                    _selectedCell = (y + _offsetY, x + _offsetX);

                    VisibleInfo = InfoType.General;

                    InfoName = "Üres";
                    InfoImage = NothingImage;

                    OnPropertyChanged(nameof(VisibleInfo));
                    OnPropertyChanged(nameof(InfoName));
                    OnPropertyChanged(nameof(InfoImage));
                }
            }


            RefreshTable();
        }

        /// <summary>
        /// Set the text of the step game button to the appropriate string
        /// </summary>
        private void SetStepGameButtonText()
        {
            switch (_model.Table.State)
            {
                case GameState.Blue:
                    StepGameButtonText = "Kör átadása";
                    break;
                case GameState.Red:
                    StepGameButtonText = "Támadás indítása";
                    break;
                case GameState.Simulation:
                    StepGameButtonText = "Támadás";
                    break;
            }

            OnPropertyChanged(nameof(StepGameButtonText));
        }

        #endregion

        #region Game event handlers

        /// <summary>
        /// Game creation event handler for the model
        /// </summary>
        private void Model_GameCreated(object sender, GameEventArgs e)
        {
            VisibleInfo = InfoType.Nothing;
            OnPropertyChanged(nameof(VisibleInfo));
            _selectedObject = SelectedObject.Nothing;

            if (_model.Table.State == GameState.MapEditor)
            {
                SelectedItem = "Üres cella";
            }
            else
            {
                SelectedItem = "-";
            }
            OnPropertyChanged(nameof(SelectedItem));

            SetStepGameButtonText();

            GenerateTiles();
        }

        /// <summary>
        /// Game advancement event handler for the model
        /// </summary>
        private void Model_GameAdvanced(object sender, GameEventArgs e)
        {
            if (e.SoldTower)
            {
                VisibleInfo = InfoType.Nothing;
                _selectedCell = (-1, -1);
                _selectedTower = (-1, -1);

                OnPropertyChanged(nameof(VisibleInfo));
            }

            if (e.UpgradedTower)
            {
                SelectCell(e.SelectedCellY - _offsetY, e.SelectedCellX - _offsetX);
            }

            if (e.WhyIsNothingHappening != null)
            {
                VisibleInfo = InfoType.Text;
                WhyIsNothingHappening = e.WhyIsNothingHappening;
                
                OnPropertyChanged(nameof(VisibleInfo));
                OnPropertyChanged(nameof(WhyIsNothingHappening));
            }

            SetStepGameButtonText();

            RefreshTable();
        }



        #endregion

        #region Event methods

        /// <summary>
        /// Instantiate new game event
        /// </summary>
        private void OnNewGame()
        {
            _selectedTower = (-1, -1);
            if (NewGame != null)
                NewGame(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate map creation event
        /// </summary>
        private void OnCreateMap()
        {
            _selectedTower = (-1, -1);
            if (CreateMap != null)
                CreateMap(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate load game event
        /// </summary>
        private void OnLoadGame()
        {
            _selectedTower = (-1, -1);
            if (LoadGame != null)
                LoadGame(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate save game event
        /// </summary>
        private void OnSaveGame()
        {
            _selectedTower = (-1, -1);
            if (SaveGame != null)
                SaveGame(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate exit game event
        /// </summary>
        private void OnExitGame()
        {
            _selectedTower = (-1, -1);
            if (ExitGame != null)
                ExitGame(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate show help event
        /// </summary>
        private void OnHelpRequested()
        {
            if (Help != null)
                Help(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate show helptower event
        /// </summary>
        private void OnHelpRequestedTower()
        {
            if (HelpTower != null)
                HelpTower(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate show helpunit event
        /// </summary>
        private void OnHelpRequestedUnit()
        {
            if (HelpUnit != null)
                HelpUnit(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate show helptower event
        /// </summary>
        private void OnHelpRequestedMapeditor()
        {
            if (HelpMapeditor != null)
                HelpMapeditor(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate open map editor event
        /// </summary>
        private void OnEditMap()
        {
            _selectedTower = (-1, -1);
            if (EditMap != null)
                EditMap(this, EventArgs.Empty);
        }

        /// <summary>
        /// Instantiate move camera up event
        /// </summary>
        public void OnCameraUp()
        {
            if (_offsetY > 0)
            {
                _selectedTower = (-1, -1);
                --_offsetY;
                if (_offsetY == 0)
                {
                    CanGoUp = false;
                }
                CanGoDown = true;
                OnPropertyChanged(nameof(CanGoUp));
                OnPropertyChanged(nameof(CanGoDown));
                RefreshTable();
            }
        }

        /// <summary>
        /// Instantiate move camera left event
        /// </summary>
        public void OnCameraLeft()
        {
            if (_offsetX > 0)
            {
                _selectedTower = (-1, -1);
                --_offsetX;
                if (_offsetX == 0)
                {
                    CanGoLeft = false;
                }
                CanGoRight = true;
                OnPropertyChanged(nameof(CanGoLeft));
                OnPropertyChanged(nameof(CanGoRight));
                RefreshTable();
            }
        }

        /// <summary>
        /// Instantiate move camera right event
        /// </summary>
        public void OnCameraRight()
        {
            if (_offsetX < _model.Table.Width - GridWidth)
            {
                _selectedTower = (-1, -1);
                ++_offsetX;
                if (_offsetX == _model.Table.Width - GridWidth)
                {
                    CanGoRight = false;
                }
                CanGoLeft = true;
                OnPropertyChanged(nameof(CanGoRight));
                OnPropertyChanged(nameof(CanGoLeft));
                RefreshTable();
            }
        }

        /// <summary>
        /// Instantiate move camera down event
        /// </summary>
        public void OnCameraDown()
        {
            if (_offsetY < _model.Table.Height - GridHeight)
            {
                _selectedTower = (-1, -1);
                ++_offsetY;
                if (_offsetY == _model.Table.Height - GridHeight)
                {
                    CanGoDown = false;
                }
                CanGoUp = true;
                OnPropertyChanged(nameof(CanGoDown));
                OnPropertyChanged(nameof(CanGoUp));
                RefreshTable();
            }
        }

        /// <summary>
        /// Selection of menu item from one of the menus
        /// </summary>
        /// <param name="menuItem">The type selected item</param>
        private void OnSelectMenuItem(SelectedObject menuItem)
        {
            if (menuItem == SelectedObject.SmallUnit || menuItem == SelectedObject.LargeUnit)
            {
                _model.BuyUnit(menuItem);
                _selectedObject = SelectedObject.Nothing;
            }
            else if (_selectedObject == menuItem)
            {
                SelectedItem = _model.Table.State == GameState.MapEditor ? "Üres cella" : "-";
                _selectedObject = SelectedObject.Nothing;
            }
            else
            {
                _selectedObject = menuItem;

                switch (menuItem)
                {
                    case SelectedObject.SmallTower:
                        SelectedItem = "Kis torony";
                        break;
                    case SelectedObject.MediumTower:
                        SelectedItem = "Közepes torony";
                        break;
                    case SelectedObject.LargeTower:
                        SelectedItem = "Nagy torony";
                        break;
                    case SelectedObject.Dune:
                        SelectedItem = "Dűne";
                        break;
                    case SelectedObject.Oasis:
                        SelectedItem = "Oázis";
                        break;
                    case SelectedObject.Ruins:
                        SelectedItem = "Rom";
                        break;
                    case SelectedObject.Castle:
                        SelectedItem = "Kastély";
                        break;
                    case SelectedObject.Barrack:
                        SelectedItem = "Barakk";
                        break;
                }
            }

            OnPropertyChanged(nameof(SelectedItem));
        }

        /// <summary>
        /// Tower selling command handler
        /// </summary>
        private void OnSellTower()
        {
            _selectedTower = (-1, -1);
            if (_selectedCell != (-1, -1))
            {
                _model.SellTower(_selectedCell.Y, _selectedCell.X);
            }
        }

        /// <summary>
        /// Tower upgrading command handler
        /// </summary>
        private void OnUpgradeTower()
        {
            if (_selectedCell != (-1, -1))
            {
                _model.UpgradeTower(_selectedCell.Y, _selectedCell.X);

                object cell = _model.Table[_selectedCell.Y, _selectedCell.X ];

                TowerDamage = ((Tower)cell).Damage;
                TowerRange = (float)Math.Round(((Tower)cell).Range, 2);
                TowerLevel = ((Tower)cell).Level;
                TowerSellCost = ((Tower)cell).SellCost;
                TowerUpgradeCost = ((Tower)cell).UpgradeCost;

                if (cell is SmallTower t)
                {
                    InfoName = "Kis torony";
                    InfoImage = t.IsBlue ? SmallTowerImageBlue : SmallTowerImageRed;
                }
                else if (cell is MediumTower m)
                {
                    InfoName = "Közepes torony";
                    InfoImage = m.IsBlue ? MediumTowerImageBlue : MediumTowerImageRed;
                }
                else if (cell is LargeTower l)
                {
                    InfoName = "Nagy torony";
                    InfoImage = l.IsBlue ? LargeTowerImageBlue : LargeTowerImageRed;
                }

                OnPropertyChanged(nameof(InfoImage));
                OnPropertyChanged(nameof(TowerDamage));
                OnPropertyChanged(nameof(TowerRange));
                OnPropertyChanged(nameof(TowerLevel));
                OnPropertyChanged(nameof(TowerSellCost));
                OnPropertyChanged(nameof(TowerUpgradeCost));
                OnPropertyChanged(nameof(InfoName));
                OnPropertyChanged(nameof(InfoImage));
            }
        }

        /// <summary>
        /// Game state stepping command handler
        /// </summary>
        private void OnStepGameState()
        {
            _selectedTower = (-1, -1);
            _model.StepGameState();
        }        

        #endregion
    }
}
